from typing import List

from cg.store import Store
from cg.store.filters.status_bed_version_filters import (
    get_bed_version_by_file_name,
    get_bed_version_by_short_name,
)
from cg.store.models import BedVersion


def test_get_bed_version_by_file_name(base_store: Store, bed_version_file_name: str):
    """Test return bed version by file name."""
    # GIVEN a store containing bed version

    # WHEN retrieving bed version
    bed_version: BedVersion = get_bed_version_by_file_name(
        bed_versions=base_store._get_query(table=BedVersion),
        bed_version_file_name=bed_version_file_name,
    ).first()

    # THEN bed version should be returned
    assert bed_version

    # THEN the file name should match the original
    assert bed_version.filename == bed_version_file_name


def test_get_bed_version_by_file_name_when_no_file(base_store: Store):
    """Test return bed version by file name when file does not exist."""
    # GIVEN a store containing bed version

    # WHEN retrieving bed version
    bed_version: BedVersion = get_bed_version_by_file_name(
        bed_versions=base_store._get_query(table=BedVersion),
        bed_version_file_name="does_not_exist",
    ).first()

    # THEN bed version should not be returned
    assert not bed_version


def test_get_bed_version_by_short_name(base_store: Store, bed_version_short_name: str):
    """Test return bed version by short name."""
    # GIVEN a store containing bed version

    # WHEN retrieving bed versions
    bed_versions: List[BedVersion] = get_bed_version_by_short_name(
        bed_versions=base_store._get_query(table=BedVersion),
        bed_version_short_name=bed_version_short_name,
    )

    # THEN bed version should be returned
    assert bed_versions

    # THEN the short name should match the original
    assert bed_versions[0].shortname == bed_version_short_name


def test_get_bed_version_by_short_name_when_no_name(base_store: Store):
    """Test return bed version by short name when short name does not exist."""
    # GIVEN a store containing bed version

    # WHEN retrieving bed versions
    bed_versions: List[BedVersion] = get_bed_version_by_short_name(
        bed_versions=base_store._get_query(table=BedVersion),
        bed_version_short_name="does_not_exist",
    )

    # THEN bed versions should not be returned
    assert not list(bed_versions)



from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, List
from unittest import mock

import pytest
from housekeeper.store.models import File
from requests import Response

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.archiving import ArchiveLocations
from cg.constants.constants import FileFormat
from cg.constants.subject import Gender
from cg.io.controller import WriteStream
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.archive.ddn_dataflow import (
    ROOT_TO_TRIM,
    DDNDataFlowClient,
    MiriaObject,
    TransferPayload,
)
from cg.meta.archive.models import FileAndSample
from cg.models.cg_config import DataFlowConfig
from cg.store import Store
from cg.store.models import Customer, Sample
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="ddn_dataflow_config")
def ddn_dataflow_config(
    local_storage_repository: str, remote_storage_repository: str
) -> DataFlowConfig:
    """Returns a mock DDN Dataflow config."""
    return DataFlowConfig(
        database_name="test_db",
        user="test_user",
        password="DummyPassword",
        url=Path("some", "api", "url.com").as_posix(),
        local_storage=local_storage_repository,
        archive_repository=remote_storage_repository,
    )


@pytest.fixture(name="ok_ddn_response")
def ok_ddn_response(ok_response: Response):
    ok_response._content = b'{"job_id": "123"}'
    return ok_response


@pytest.fixture(name="archive_request_json")
def archive_request_json(
    remote_storage_repository: str, local_storage_repository: str, trimmed_local_path: str
) -> Dict:
    return {
        "osType": "Unix/MacOS",
        "createFolder": False,
        "pathInfo": [
            {
                "destination": f"{remote_storage_repository}ADM1",
                "source": local_storage_repository + trimmed_local_path,
            }
        ],
        "metadataList": [],
    }


@pytest.fixture(name="retrieve_request_json")
def retrieve_request_json(
    remote_storage_repository: str, local_storage_repository: str, trimmed_local_path: str
) -> Dict[str, Any]:
    """Returns the body for a retrieval http post towards the DDN Miria API."""
    return {
        "osType": "Unix/MacOS",
        "createFolder": False,
        "pathInfo": [
            {
                "destination": local_storage_repository
                + Path(trimmed_local_path).parent.as_posix(),
                "source": f"{remote_storage_repository}ADM1",
            }
        ],
        "metadataList": [],
    }


@pytest.fixture(name="header_with_test_auth_token")
def header_with_test_auth_token() -> Dict:
    return {
        "Content-Type": "application/json",
        "accept": "application/json",
        "Authorization": "Bearer test_auth_token",
    }


@pytest.fixture(name="ddn_auth_token_response")
def ddn_auth_token_response(ok_response: Response):
    ok_response._content = b'{"access": "test_auth_token", "expire":15, "test_refresh_token"}'
    return ok_response


@pytest.fixture(name="ddn_dataflow_client")
def ddn_dataflow_client(ddn_dataflow_config: DataFlowConfig) -> DDNDataFlowClient:
    """Returns a DDNApi without tokens being set."""
    mock_ddn_auth_success_response = Response()
    mock_ddn_auth_success_response.status_code = 200
    mock_ddn_auth_success_response._content = WriteStream.write_stream_from_content(
        file_format=FileFormat.JSON,
        content={
            "access": "test_auth_token",
            "refresh": "test_refresh_token",
            "expire": int((datetime.now() + timedelta(minutes=20)).timestamp()),
        },
    ).encode()
    with mock.patch(
        "cg.meta.archive.ddn_dataflow.APIRequest.api_request_from_content",
        return_value=mock_ddn_auth_success_response,
    ):
        return DDNDataFlowClient(ddn_dataflow_config)


@pytest.fixture(name="miria_file_archive")
def miria_file(local_directory: Path, remote_path: Path) -> MiriaObject:
    """Return a MiriaObject for archiving."""
    return MiriaObject(source=local_directory.as_posix(), destination=remote_path.as_posix())


@pytest.fixture(name="file_and_sample")
def file_and_sample(spring_archive_api: SpringArchiveAPI, sample_id: str):
    return FileAndSample(
        file=spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first(),
        sample=spring_archive_api.status_db.get_sample_by_internal_id(sample_id),
    )


@pytest.fixture(name="trimmed_local_path")
def trimmed_local_path(spring_archive_api: SpringArchiveAPI, sample_id: str):
    file: File = spring_archive_api.housekeeper_api.get_files(bundle=sample_id).first()
    return file.path[5:]


@pytest.fixture(name="miria_file_retrieve")
def miria_file_retrieve(local_directory: Path, remote_path: Path) -> MiriaObject:
    """Return a MiriaObject for retrieval."""
    return MiriaObject(source=remote_path.as_posix(), destination=local_directory.as_posix())


@pytest.fixture(name="transfer_payload")
def transfer_payload(miria_file_archive: MiriaObject) -> TransferPayload:
    """Return a TransferPayload object containing two identical MiriaObject object."""
    return TransferPayload(
        files_to_transfer=[miria_file_archive, miria_file_archive.model_copy(deep=True)]
    )


@pytest.fixture(name="remote_path")
def remote_path() -> Path:
    """Returns a mock path."""
    return Path("/some", "place")


@pytest.fixture(name="local_directory")
def local_directory() -> Path:
    """Returns a mock path with /home as its root."""
    return Path(ROOT_TO_TRIM, "other", "place")


@pytest.fixture(name="trimmed_local_directory")
def trimmed_local_directory(local_directory: Path) -> Path:
    """Returns the trimmed local directory."""
    return Path(f"/{local_directory.relative_to(ROOT_TO_TRIM)}")


@pytest.fixture(name="local_storage_repository")
def local_storage_repository() -> str:
    """Returns a local storage repository."""
    return "local@storage:"


@pytest.fixture(name="remote_storage_repository")
def remote_storage_repository() -> str:
    """Returns a remote storage repository."""
    return "archive@repository:"


@pytest.fixture(name="full_remote_path")
def full_remote_path(remote_storage_repository: str, remote_path: Path) -> str:
    """Returns the merged remote repository and path."""
    return remote_storage_repository + remote_path.as_posix()


@pytest.fixture(name="archive_store")
def archive_store(
    base_store: Store,
    helpers: StoreHelpers,
    sample_id,
    father_sample_id,
    mother_sample_id,
    sample_name,
) -> Store:
    """Returns a store with samples for both a DDN customer and a non-DDN customer."""
    customer_ddn: Customer = base_store.add_customer(
        internal_id="CustWithDDN",
        invoice_address="Baker Street 221B",
        invoice_reference="Sherlock Holmes",
        name="Sherlock Holmes",
        is_clinical=True,
        data_archive_location=ArchiveLocations.KAROLINSKA_BUCKET,
    )
    customer_without_ddn: Customer = base_store.add_customer(
        internal_id="CustWithoutDDN",
        invoice_address="Wallaby Way 42",
        invoice_reference="P Sherman",
        name="P Sherman",
        is_clinical=False,
        data_archive_location="PDC",
    )

    new_samples: List[Sample] = [
        base_store.add_sample(
            name=sample_name,
            sex=Gender.MALE,
            internal_id=sample_id,
        ),
        base_store.add_sample(
            name="sample_2_with_ddn_customer",
            sex=Gender.MALE,
            internal_id=mother_sample_id,
        ),
        base_store.add_sample(
            name="sample_without_ddn_customer",
            sex=Gender.MALE,
            internal_id=father_sample_id,
        ),
    ]
    new_samples[0].customer = customer_ddn
    new_samples[1].customer = customer_ddn
    new_samples[2].customer = customer_without_ddn
    external_app = base_store.get_application_by_tag("WGXCUSC000").versions[0]
    wgs_app = base_store.get_application_by_tag("WGSPCFC030").versions[0]
    for sample in new_samples:
        sample.application_version = external_app if "external" in sample.name else wgs_app
    base_store.session.add(customer_ddn)
    base_store.session.add(customer_without_ddn)
    base_store.session.add_all(new_samples)
    base_store.session.commit()
    return base_store


@pytest.fixture(name="sample_with_spring_file")
def sample_with_spring_file() -> str:
    return "ADM1"


@pytest.fixture(name="spring_archive_api")
def spring_archive_api(
    populated_housekeeper_api: HousekeeperAPI,
    archive_store: Store,
    ddn_dataflow_config: DataFlowConfig,
    father_sample_id: str,
    helpers,
) -> SpringArchiveAPI:
    """Returns an ArchiveAPI with a populated housekeeper store and a DDNDataFlowClient.
    Also adds /home/ as a prefix for each SPRING file for the DDNDataFlowClient to accept them."""
    for spring_file in populated_housekeeper_api.files(tags=[SequencingFileTag.SPRING]):
        spring_file.path = f"/home/{spring_file.path}"
    populated_housekeeper_api.commit()
    return SpringArchiveAPI(
        housekeeper_api=populated_housekeeper_api,
        status_db=archive_store,
        data_flow_config=ddn_dataflow_config,
    )



from datetime import datetime
from typing import Optional, Union

from pydantic.v1 import BaseModel, validator

from cg.constants.subject import Gender
from cg.models.report.metadata import SampleMetadataModel
from cg.models.report.validators import (
    validate_boolean,
    validate_date,
    validate_empty_field,
    validate_gender,
    validate_rml_sample,
)


class ApplicationModel(BaseModel):
    """
    Analysis application attributes model

    Attributes:
        tag: application identifier; source: StatusDB/application/tag
        version: application version; source: LIMS/sample/application_version
        prep_category: library preparation category; source: StatusDB/application/prep_category
        description: analysis description; source: StatusDB/application/description
        limitations: application limitations; source: StatusDB/application/limitations
        accredited: if the sample associated process is accredited or not; source: StatusDB/application/is_accredited
        external: whether the app tag is external or not; source: StatusDB/application/is_external
    """

    tag: Optional[str]
    version: Union[None, int, str]
    prep_category: Optional[str]
    description: Optional[str]
    limitations: Optional[str]
    accredited: Optional[bool]
    external: Optional[bool]

    _prep_category = validator("prep_category", always=True, allow_reuse=True)(validate_rml_sample)
    _values = validator(
        "tag",
        "version",
        "prep_category",
        "description",
        "limitations",
        always=True,
        allow_reuse=True,
    )(validate_empty_field)


class MethodsModel(BaseModel):
    """
    Model describing the methods used for preparation and sequencing of the samples

    Attributes:
        library_prep: library preparation method; source: LIMS/sample/prep_method
        sequencing: sequencing procedure; source: LIMS/sample/sequencing_method
    """

    library_prep: Optional[str]
    sequencing: Optional[str]

    _values = validator("library_prep", "sequencing", always=True, allow_reuse=True)(
        validate_empty_field
    )


class TimestampModel(BaseModel):
    """
    Model describing the processing timestamps of a specific sample

    Atributes:
        ordered_at: order date; source: StatusDB/sample/ordered_at
        received_at: arrival date; source: StatusDB/sample/received_at
        prepared_at: library preparation date; source: StatusDB/sample/prepared_at
        reads_updated_at: sequencing date; source: StatusDB/sample/reads_updated_at
    """

    ordered_at: Union[None, datetime, str]
    received_at: Union[None, datetime, str]
    prepared_at: Union[None, datetime, str]
    reads_updated_at: Union[None, datetime, str]

    _values = validator(
        "ordered_at",
        "received_at",
        "prepared_at",
        "reads_updated_at",
        always=True,
        allow_reuse=True,
    )(validate_date)


class SampleModel(BaseModel):
    """
    Sample attributes model

    Attributes:
        name: sample name; source: StatusDB/sample/name
        id: sample internal ID; source: StatusDB/sample/internal_id
        ticket: ticket number; source: StatusDB/sample/ticket_number
        status: sample status provided by the customer; source: StatusDB/family-sample/status
        gender: sample gender provided by the customer; source: StatusDB/sample/sex
        source: sample type/source; source: LIMS/sample/source
        tumour: whether the sample is a tumour or normal one; source: StatusDB/sample/is_tumour
        application: analysis application model
        methods: sample processing methods model
        metadata: sample associated metrics and trending data model
        timestamps: processing timestamp attributes
    """

    name: Optional[str]
    id: Optional[str]
    ticket: Union[None, int, str]
    status: Optional[str]
    gender: Optional[str] = Gender.UNKNOWN
    source: Optional[str]
    tumour: Union[None, bool, str]
    application: ApplicationModel
    methods: MethodsModel
    metadata: SampleMetadataModel
    timestamps: TimestampModel

    _tumour = validator("tumour", always=True, allow_reuse=True)(validate_boolean)
    _gender = validator("gender", always=True, allow_reuse=True)(validate_gender)
    _values = validator("name", "id", "ticket", "status", "source", always=True, allow_reuse=True)(
        validate_empty_field
    )



"""Tests for the functions in the store api status related to the customer module."""
from typing import List

from cg.constants.invoice import CustomerNames
from cg.store import Store
from cg.store.models import Customer
from tests.store_helpers import StoreHelpers


def test_get_customers_to_invoice(
    store_with_samples_for_multiple_customers: Store,
    helpers: StoreHelpers,
    three_customer_ids: List[str],
):
    """Test that customers to invoice can be returned."""
    # GIVEN a database with samples for a customer

    # WHEN getting the customers to invoice
    customers: List[Customer] = store_with_samples_for_multiple_customers.get_customers_to_invoice(
        records=store_with_samples_for_multiple_customers.get_samples_to_invoice_query()
    )

    # THEN the customers should be returned
    assert customers
    assert len(customers) == 2
    for customer in customers:
        assert customer.internal_id in three_customer_ids

    # THEN the customers to invoice should not contain cust000
    for customer in customers:
        assert customer.internal_id != CustomerNames.CG_INTERNAL_CUSTOMER



from typing import Dict

import pytest

from cg.constants import METAGENOME_SOURCES
from cg.constants.orderforms import REV_SEX_MAP
from cg.models.orders.constants import ExcelSampleAliases
from cg.models.orders.excel_sample import ExcelSample
from cg.models.orders.sample_base import PriorityEnum


def test_collection_date_conversion(mip_rna_orderform_sample: Dict):
    """Tests that a sample with Collection Date set to YYYY-MM-DD HH:MM:SS is converted to YYYY-MM-DD."""

    # GIVEN a parsed order form in Excel format with Collection Date set
    mip_rna_orderform_sample[ExcelSampleAliases.COLLECTION_DATE] = "2023-09-29 00:00:00"

    # WHEN converting the sample to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the date should have been converted
    assert excel_sample.collection_date == "2023-09-29"


def test_collection_date_none(mip_rna_orderform_sample: Dict):
    """Tests that a sample with no Collection Date set is allowed."""

    # GIVEN a parsed order form in Excel format with no Collection Date set (Note that these are parsed as empty strings
    # rather than None
    mip_rna_orderform_sample[ExcelSampleAliases.COLLECTION_DATE] = ""

    # THEN converting the sample to an ExcelSample object should not raise an error
    assert ExcelSample.model_validate(mip_rna_orderform_sample)


def test_concentration_numeric(mip_rna_orderform_sample: Dict):
    """Tests that a sample with concentration set to a natural number is allowed."""

    # GIVEN a parsed order form in Excel format with concentration set to a natural number
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION] = "2"

    # THEN converting the sample to an ExcelSample object should not raise an error
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    assert excel_sample.concentration == "2"


def test_concentration_fraction(mip_rna_orderform_sample: Dict):
    """Tests that a sample with concentration set as a fraction is allowed."""

    # GIVEN a parsed order form in Excel format with concentration set to a fraction
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION] = "0.23"

    # THEN converting the sample to an ExcelSample object should not raise an error
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the parsed ExcelSample object should have its concentration set
    assert excel_sample.concentration == "0.23"


def test_concentration_numeric_fraction(mip_rna_orderform_sample: Dict):
    """Tests that a sample with numeric concentration set as a float is allowed and converted."""

    # GIVEN a parsed order form in Excel format with concentration set as a float with only zeroes
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION] = "2.00"

    # THEN converting the sample to an ExcelSample object should not raise an error
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the parsed ExcelSample object should have its concentration set without zeroes
    assert excel_sample.concentration == "2"


def test_concentration_alphabetical_fail(mip_rna_orderform_sample: Dict):
    """Tests that a sample with non-numeric concentration raises an AttributeError."""

    # GIVEN a parsed order form in Excel format with concentration set with alphabetical characters
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION] = "2 nM"

    # THEN converting the sample to an ExcelSample object should raise an AttributeError
    with pytest.raises(AttributeError):
        ExcelSample.model_validate(mip_rna_orderform_sample)


def test_concentration_sample_numeric(mip_rna_orderform_sample: Dict):
    """Tests that a sample with concentration_sample set to a natural number is allowed."""

    # GIVEN a parsed order form in Excel format with sample concentration set to a natural number
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION_SAMPLE] = "2"

    # THEN converting the sample to an ExcelSample object should not raise an error
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the ExcelSample object should have its concentration sample set to 2
    assert excel_sample.concentration_sample == "2"


def test_concentration_sample_fraction(mip_rna_orderform_sample: Dict):
    """Tests that a sample with concentration sample set as a fraction is allowed."""

    # GIVEN a parsed order form in Excel format with concentration sample set to a fraction
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION_SAMPLE] = "0.23"

    # THEN converting the sample to an ExcelSample object should not raise an error
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the parsed ExcelSample object should have its concentration sample set
    assert excel_sample.concentration_sample == "0.23"


def test_concentration_sample_numeric_fraction(mip_rna_orderform_sample: Dict):
    """Tests that a sample with numeric concentration sample set as a float is allowed and converted."""

    # GIVEN a parsed order form in Excel format with concentration sample set as a float with only zeroes
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION_SAMPLE] = "2.00"

    # THEN converting the sample to an ExcelSample object should not raise an error
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the parsed ExcelSample object should have its concentration sample set without zeroes
    assert excel_sample.concentration_sample == "2"


def test_concentration_sample_alphabetical_fail(mip_rna_orderform_sample: Dict):
    """Tests that a sample with non-numeric concentration sample raises an AttributeError."""

    # GIVEN a parsed order form in Excel format with concentration sample set with alphabetical characters
    mip_rna_orderform_sample[ExcelSampleAliases.CONCENTRATION_SAMPLE] = "2 nM"

    # THEN converting the sample to an ExcelSample object should raise an AttributeError
    with pytest.raises(AttributeError):
        ExcelSample.model_validate(mip_rna_orderform_sample)


def test_data_analysis_valid(mip_rna_orderform_sample: Dict):
    """Tests that a sample with a valid data analysis set is allowed."""

    # GIVEN a parsed order form in Excel format with data analysis set to MIP RNA

    # THEN no error should be raised when converting to an ExcelSample object
    assert ExcelSample.model_validate(mip_rna_orderform_sample)


def test_data_analysis_invalid(mip_rna_orderform_sample: Dict):
    """Tests that a sample with an invalid data analysis set is not allowed."""

    # GIVEN a parsed order form in Excel format with data analysis set to an invalid value
    mip_rna_orderform_sample[ExcelSampleAliases.DATA_ANALYSIS] = "INVALID ANALYSIS"

    # THEN an AttributeError should be raised when converting to an ExcelSample object
    with pytest.raises(AttributeError):
        ExcelSample.model_validate(mip_rna_orderform_sample)


def test_data_delivery_convert_to_lower(mip_rna_orderform_sample: Dict):
    """Tests that a sample with a data delivery which is not all lower case is converted correctly."""

    # GIVEN a parsed order form in Excel format with data delivery set in all caps
    mip_rna_orderform_sample[ExcelSampleAliases.DATA_DELIVERY] = "ANALYSIS"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the data delivery should be in lower case
    assert excel_sample.data_delivery == "analysis"


def test_father_convert_zeroes(mip_rna_orderform_sample: Dict):
    """Tests that if 0.0 is provided as father, then None is set instead."""

    # GIVEN a parsed order form in Excel format with 'father' set to 0.0
    mip_rna_orderform_sample[ExcelSampleAliases.FATHER] = "0.0"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN 'father' should be set to None
    assert excel_sample.father is None


def test_parse_panels_none(mip_rna_orderform_sample: Dict):
    """Tests that the UDF/Gene List is allowed to be none for a sample in an order form."""

    # GIVEN a parsed order form in Excel format with no panels set (Note that these are parsed as empty strings
    # rather than None
    mip_rna_orderform_sample[ExcelSampleAliases.PANELS] = ""

    # THEN the model validation should succeed
    ExcelSample.model_validate(mip_rna_orderform_sample)


def test_parse_panels_semicolon_separator(mip_rna_orderform_sample: Dict):
    """Tests that the panel list is split when given with a semicolon separated string."""

    # GIVEN a parsed order form in Excel format with panels separated with a semicolon
    mip_rna_orderform_sample[ExcelSampleAliases.PANELS] = "gene_1;gene_2"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    assert excel_sample.panels == ["gene_1", "gene_2"]


def test_parse_panels_colon_separator(mip_rna_orderform_sample: Dict):
    """Tests that the panel list is split when given with a colon separated string."""

    # GIVEN a parsed order form in Excel format with panels separated with a colon
    mip_rna_orderform_sample[ExcelSampleAliases.PANELS] = "gene_1:gene_2"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    assert excel_sample.panels == ["gene_1", "gene_2"]


def test_priority_foertur(mip_rna_orderform_sample: Dict):
    """Tests that a sample with priority specified to 'förtur', then the validation converts the value to English."""

    # GIVEN a parsed order form in Excel format with priority set to 'förtur'
    mip_rna_orderform_sample[ExcelSampleAliases.PRIORITY] = "förtur"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the priority should be set to 'priority'
    assert excel_sample.priority == PriorityEnum.priority


def test_valid_priority_all_caps(mip_rna_orderform_sample: Dict):
    """Tests that a sample with a valid priority specified in all caps is converted to lower case."""

    # GIVEN a parsed order form in Excel format with priority set to "PRIORITY"
    mip_rna_orderform_sample[ExcelSampleAliases.PRIORITY] = "PRIORITY"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the priority should be in lower case
    assert excel_sample.priority == PriorityEnum.priority


def test_priority_foertur_all_caps(mip_rna_orderform_sample: Dict):
    """Tests that a sample with priority specified to 'FÖRTUR',
    then the validation converts the value to English and lower case."""

    # GIVEN a parsed order form in Excel format with priority set to 'FÖRTUR'
    mip_rna_orderform_sample[ExcelSampleAliases.PRIORITY] = "FÖRTUR"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the priority should be set to 'priority'
    assert excel_sample.priority == PriorityEnum.priority


def test_valid_priority_with_spaces(mip_rna_orderform_sample: Dict):
    """Tests that a sample with a valid priority has its spaces replaced by underscores."""

    # GIVEN a parsed order form in Excel format with priority set to "Clinical trials"
    mip_rna_orderform_sample[ExcelSampleAliases.PRIORITY] = "Clinical trials"

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the priority should set to "clinical_trials"
    assert excel_sample.priority == PriorityEnum.clinical_trials


def test_convert_sex_m(mip_rna_orderform_sample: Dict):
    """Tests that sex is converted to 'male' if set as 'M'."""

    # GIVEN a parsed order form in Excel format with sex set to 'M'

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the sex should be set to 'male'
    assert excel_sample.sex == REV_SEX_MAP["M"]


def test_convert_sex_strip(mip_rna_orderform_sample: Dict):
    """Tests that sex is stripped from any trailing spaces."""

    # GIVEN a parsed order form in Excel format with sex set to 'M '
    mip_rna_orderform_sample[ExcelSampleAliases.SEX] = "M "

    # WHEN converting to an ExcelSample object
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the sex should have been converted to 'male'
    assert excel_sample.sex == REV_SEX_MAP["M"]


def test_invalid_source(mip_rna_orderform_sample: Dict):
    """Tests that a ValueError is raised when an invalid source is specified."""

    # GIVEN a parsed order form in Excel format with an invalid source specified
    mip_rna_orderform_sample[ExcelSampleAliases.SOURCE] = "INVALID SOURCE"

    # WHEN converting to an ExcelSample object

    # THEN a ValueError should be raised
    with pytest.raises(ValueError):
        ExcelSample.model_validate(mip_rna_orderform_sample)


def test_valid_source(mip_rna_orderform_sample: Dict):
    """Tests that no ValueError is raised when a valid source is specified."""

    # GIVEN a parsed order form in Excel format with a valid source specified
    mip_rna_orderform_sample[ExcelSampleAliases.SOURCE] = METAGENOME_SOURCES[0]

    # WHEN converting to an ExcelSample object

    # THEN no ValueError should be raised
    ExcelSample.model_validate(mip_rna_orderform_sample)


def test_status_convert_to_lower(mip_rna_orderform_sample: Dict):
    """Tests that the given status value is converted to lower case."""

    # GIVEN a parsed order form in Excel format with a valid source specified
    mip_rna_orderform_sample[ExcelSampleAliases.STATUS] = "ANALYSIS"

    # WHEN converting to an ExcelSample object

    # THEN no ValueError should be raised
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the status should be in lower case
    assert excel_sample.status == "analysis"


def test_volume_non_numeric(mip_rna_orderform_sample: Dict):
    """Tests that an error is raised when specifying a non-numeric volume."""

    # GIVEN a parsed order form in Excel format with non-numeric volume specified
    mip_rna_orderform_sample[ExcelSampleAliases.VOLUME] = "2 nM"

    # WHEN converting to an ExcelSample object

    # THEN an AttributeError should be raised
    with pytest.raises(AttributeError):
        ExcelSample.model_validate(mip_rna_orderform_sample)


def test_volume_positive_integer(mip_rna_orderform_sample: Dict):
    """Tests that no error is raised when parsing a volume set to a positive integer."""

    # GIVEN a parsed order form in Excel format with a positive integer volume
    mip_rna_orderform_sample[ExcelSampleAliases.VOLUME] = "2"

    # WHEN converting to an ExcelSample object

    # THEN no AttributeError should be raised
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the volume should be unchanged
    assert excel_sample.volume == "2"


def test_volume_decimal_integer(mip_rna_orderform_sample: Dict):
    """Tests that an integer given in float format, i.e. 2.0 instead of 2 and so on, is converted to integer format."""

    # GIVEN a parsed order form in Excel format with an integer volume in float format
    mip_rna_orderform_sample[ExcelSampleAliases.VOLUME] = "2.0"

    # WHEN converting to an ExcelSample object

    # THEN no AttributeError should be raised
    excel_sample: ExcelSample = ExcelSample.model_validate(mip_rna_orderform_sample)

    # THEN the volume should be converted to have no decimals
    assert excel_sample.volume == "2"



import logging
from pathlib import Path
from typing import List

import pytest

from cg.constants.demultiplexing import RunParametersXMLNodes
from cg.constants.sequencing import Sequencers
from cg.exc import RunParametersError
from cg.models.demultiplex.run_parameters import (
    RunParameters,
    RunParametersNovaSeq6000,
    RunParametersNovaSeqX,
)


def test_run_parameters_parent_class_fails(
    novaseq_6000_run_parameters_path: Path,
    novaseq_x_run_parameters_path: Path,
):
    """Test that trying to instantiate the parent RunParameters class raises an error."""
    # GIVEN valid paths for run parameters files

    # WHEN trying to instantiate the parent class with a NovaSeq6000 file
    with pytest.raises(NotImplementedError) as exc_info:
        # THEN an NotImplementedError is raised
        RunParameters(run_parameters_path=novaseq_6000_run_parameters_path)
        assert (
            str(exc_info.value)
            == "Parent class instantiated. Please instantiate a child class instead"
        )

    # WHEN trying to instantiate the parent class with a NovaSeqX file
    with pytest.raises(NotImplementedError) as exc_info:
        # THEN an NotImplementedError is raised
        RunParameters(run_parameters_path=novaseq_x_run_parameters_path)
        assert (
            str(exc_info.value)
            == "Parent class instantiated. Please instantiate a child class instead"
        )


def test_run_parameters_novaseq_6000(novaseq_6000_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeq6000 object with the correct file works."""

    # GIVEN a valid NovaSeq6000 run parameters file path

    # WHEN creating a NovaSeq6000 RunParameters object
    run_parameters = RunParametersNovaSeq6000(run_parameters_path=novaseq_6000_run_parameters_path)

    # THEN the created object is of the correct type and has the correct attributes
    assert isinstance(run_parameters, RunParametersNovaSeq6000)
    assert run_parameters.sequencer == Sequencers.NOVASEQ.value


def test_run_parameters_novaseq_6000_wrong_file(novaseq_x_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeq6000 object with the wrong file fails."""
    # GIVEN a file path with a run parameters file from an instrument different from NovaSeq6000

    # WHEN trying to create a NovaSeq6000 RunParameters object with the file
    with pytest.raises(RunParametersError) as exc_info:
        # THEN an error is raised
        RunParametersNovaSeq6000(run_parameters_path=novaseq_x_run_parameters_path)
        assert (
            str(exc_info.value) == "The file parsed does not correspond to a NovaSeq6000 instrument"
        )


def test_run_parameters_novaseq_x(novaseq_x_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeqX object with the correct file works."""

    # GIVEN a valid NovaSeqX run parameters file path

    # WHEN creating a NovaSeqX RunParameters object
    run_parameters = RunParametersNovaSeqX(run_parameters_path=novaseq_x_run_parameters_path)

    # THEN the created object is of the correct type and has the correct attributes
    assert isinstance(run_parameters, RunParametersNovaSeqX)
    assert run_parameters.sequencer == Sequencers.NOVASEQX.value


def test_run_parameters_novaseq_x_wrong_file(novaseq_6000_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeqX object with the wrong file fails."""
    # GIVEN a file path with a run parameters file from an instrument different from NovaSeqX

    # WHEN trying to create a NovaSeqX RunParameters object with the file
    with pytest.raises(RunParametersError) as exc_info:
        # THEN an error is raised
        RunParametersNovaSeqX(run_parameters_path=novaseq_6000_run_parameters_path)
        assert str(exc_info.value) == "The file parsed does not correspond to a NovaSeqX instrument"


def test_reagent_kit_version(novaseq_6000_run_parameters: RunParametersNovaSeq6000):
    """Test that getting reagent kit version from a correct file returns an expected value."""
    # GIVEN a valid RunParameters object for NovaSeq6000

    # WHEN fetching the reagent kit version
    reagent_kit_version: str = novaseq_6000_run_parameters.reagent_kit_version

    # THEN the reagent kit version exists and is not "unknown"
    assert reagent_kit_version
    assert reagent_kit_version != RunParametersXMLNodes.UNKNOWN_REAGENT_KIT_VERSION


def test_reagent_kit_version_missing_version(
    run_parameters_missing_versions: RunParametersNovaSeq6000, caplog
):
    """Test that 'unknown' will be returned if the run parameters file has no reagent kit method."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object for NovaSeq6000 created from a file without reagent kit version

    # WHEN fetching the reagent kit version
    reagent_kit: str = run_parameters_missing_versions.reagent_kit_version

    # THEN the reagent kit version is unknown
    assert reagent_kit == RunParametersXMLNodes.UNKNOWN_REAGENT_KIT_VERSION
    assert "Could not determine reagent kit version" in caplog.text


def test_reagent_kit_version_novaseq_x(novaseq_x_run_parameters: RunParametersNovaSeqX):
    """Test that getting reagent kit version from a NovaSeqX run parameters returns None."""
    # GIVEN a valid RunParameters object for NovaSeqX

    # WHEN fetching the reagent kit version

    # THEN the reagent kit version is None
    assert not novaseq_x_run_parameters.reagent_kit_version


def test_control_software_version(novaseq_6000_run_parameters: RunParametersNovaSeq6000):
    """Test that getting control software version from a correct file returns an expected value."""
    # GIVEN a valid RunParameters object for NovaSeq6000

    # WHEN fetching the control software version
    control_software_version: str = novaseq_6000_run_parameters.control_software_version

    # THEN the control software version is a non-empty string
    assert isinstance(control_software_version, str)
    assert control_software_version != ""


def test_control_software_version_no_version(run_parameters_missing_versions: Path, caplog):
    """Test that fetching the control software version from a file without that field fails."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object created from a file without control software version

    # WHEN fetching the control software version
    with pytest.raises(RunParametersError):
        # THEN assert that an exception was raised since the control software version was not found
        run_parameters_missing_versions.control_software_version

    assert "Could not determine control software version" in caplog.text


def test_control_software_version_novaseq_x(novaseq_x_run_parameters: RunParametersNovaSeqX):
    """Test that getting control software version from a NovaSeqX run parameters returns None."""
    # GIVEN a valid RunParameters object for NovaSeqX

    # WHEN fetching the control software version

    # THEN the control software version is None
    assert not novaseq_x_run_parameters.control_software_version


def test_index_length_novaseq_6000(
    novaseq_6000_run_parameters: RunParametersNovaSeq6000,
):
    """Test that getting the index length from a NovaSeq6000 run parameters file returns an int."""
    # GIVEN a valid RunParametersNovaSeq6000 object

    # WHEN getting the index length

    # THEN the index length is an int
    assert isinstance(novaseq_6000_run_parameters.index_length, int)


def test_index_length_novaseq_x(
    novaseq_x_run_parameters: RunParametersNovaSeqX,
):
    """Test that getting the index length from a NovaSeqX run parameters file returns an int."""
    # GIVEN a valid RunParametersNovaSeqX object

    # WHEN getting the index length

    # THEN the index length is an int
    assert isinstance(novaseq_x_run_parameters.index_length, int)


def test_index_length_novaseq_6000_different_length(
    run_parameters_novaseq_6000_different_index_path: Path,
    run_parameters_novaseq_x_different_index_path: Path,
):
    """Test that getting the index length from a file with different index cycles fails."""
    # GIVEN a RunParameters object created from a file with different index cycles
    run_parameters_novaseq_6000 = RunParametersNovaSeq6000(
        run_parameters_path=run_parameters_novaseq_6000_different_index_path
    )

    # WHEN fetching index length
    with pytest.raises(RunParametersError) as exc_info:
        # THEN assert that an exception was raised since the index cycles are different
        run_parameters_novaseq_6000.index_length
        assert str(exc_info.value) == "Index lengths are not the same!"


def test_index_length_novaseq_x_different_length(
    run_parameters_novaseq_x_different_index_path: Path,
):
    """Test that getting the index length from a file with different index cycles fails."""
    # GIVEN a RunParameters object created from a file with different index cycles
    run_parameters_novaseq_x = RunParametersNovaSeqX(
        run_parameters_path=run_parameters_novaseq_x_different_index_path
    )
    # WHEN fetching index length
    with pytest.raises(RunParametersError) as exc_info:
        # THEN assert that an exception was raised since the index cycles are different
        run_parameters_novaseq_x.index_length
        assert str(exc_info.value) == "Index lengths are not the same!"


def test_get_cycles_novaseq_6000(novaseq_6000_run_parameters: RunParametersNovaSeq6000):
    """Test that the read and index cycles are read correctly for NovaSeqX run parameters."""
    # GIVEN a NovaSeq6000 run parameters object

    # WHEN getting any read cycle
    read_cycles: List[int] = [
        novaseq_6000_run_parameters.get_read_1_cycles(),
        novaseq_6000_run_parameters.get_read_2_cycles(),
        novaseq_6000_run_parameters.get_index_1_cycles(),
        novaseq_6000_run_parameters.get_index_2_cycles(),
    ]

    # THEN all read cycles are non-negative integers
    for cycles in read_cycles:
        assert isinstance(cycles, int)
        assert cycles >= 0


def test_get_cycles_novaseq_x(novaseq_x_run_parameters: RunParametersNovaSeqX):
    """Test that the read and index cycles are read correctly for NovaSeqX run parameters."""
    # GIVEN a NovaSeqX run parameters object

    # WHEN getting any read cycle
    read_cycles: List[int] = [
        novaseq_x_run_parameters.get_read_1_cycles(),
        novaseq_x_run_parameters.get_read_2_cycles(),
        novaseq_x_run_parameters.get_index_1_cycles(),
        novaseq_x_run_parameters.get_index_2_cycles(),
    ]

    # THEN all read cycles are non-negative integers
    for cycles in read_cycles:
        assert isinstance(cycles, int)
        assert cycles >= 0



import json
from pathlib import Path
from typing import Dict

import pytest

from cg.constants.demultiplexing import (
    BCL2FASTQ_METRICS_DIRECTORY_NAME,
    BCL2FASTQ_METRICS_FILE_NAME,
)


@pytest.fixture
def raw_bcl2fastq_tile_metrics() -> Dict:
    """Metrics for a single tile in a lane."""
    return {
        "Flowcell": "AB1",
        "RunNumber": 1,
        "RunId": "RUN1",
        "ConversionResults": [
            {
                "LaneNumber": 1,
                "DemuxResults": [
                    {
                        "SampleId": "S1",
                        "NumberReads": 1,
                        "Yield": 100,
                        "ReadMetrics": [{"Yield": 100, "YieldQ30": 90, "QualityScoreSum": 100}],
                    }
                ],
                "Undetermined": {
                    "NumberReads": 1,
                    "Yield": 100,
                    "ReadMetrics": [
                        {
                            "ReadNumber": 1,
                            "Yield": 100,
                            "YieldQ30": 80,
                            "QualityScoreSum": 3000,
                            "TrimmedBases": 0,
                        }
                    ],
                },
            }
        ],
    }


@pytest.fixture
def bcl2fastq_flow_cell_path(
    tmp_path: Path,
    raw_bcl2fastq_tile_metrics: Dict,
) -> Path:
    """Flow cell with one sample in one lane, two tiles, two paired reads and two undetermined paired reads."""

    tile_1 = Path(tmp_path, "l1t1", BCL2FASTQ_METRICS_DIRECTORY_NAME)
    tile_1.mkdir(parents=True)
    tile_1_stats_json_path = Path(tile_1, BCL2FASTQ_METRICS_FILE_NAME)

    tile_2 = Path(tmp_path, "l1t2", BCL2FASTQ_METRICS_DIRECTORY_NAME)
    tile_2.mkdir(parents=True)
    tile_2_stats_json_path = Path(tile_2, BCL2FASTQ_METRICS_FILE_NAME)

    tile_1_stats_json_path.write_text(json.dumps(raw_bcl2fastq_tile_metrics))
    tile_2_stats_json_path.write_text(json.dumps(raw_bcl2fastq_tile_metrics))

    return tmp_path



"""Tests for the compress fastq cli."""

import datetime as dt
import logging
from typing import List

from click.testing import CliRunner

from cg.cli.compress.fastq import fastq_cmd, get_cases_to_process
from cg.constants import Pipeline
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Family
from tests.store_helpers import StoreHelpers

MOCK_SET_MEM_ACCORDING_TO_READS_PATH: str = "cg.cli.compress.helpers.set_memory_according_to_reads"


def test_get_cases_to_process(
    case_id: str,
    cli_runner: CliRunner,
    helpers: StoreHelpers,
    populated_compress_context: CGConfig,
):
    """Test get cases to process."""

    # GIVEN a populated store
    status_db: Store = populated_compress_context.status_db

    # GIVEN a context with a case that can be compressed

    valid_compressable_case: Family = helpers.add_case(
        store=status_db,
        name=case_id,
        internal_id=case_id,
        data_analysis=Pipeline.MIP_DNA,
        action=None,
    )
    valid_compressable_case.created_at = dt.datetime.now() - dt.timedelta(days=1000)
    status_db.session.commit()

    # WHEN running the compress command
    cases: List[Family] = get_cases_to_process(days_back=1, store=status_db)

    # THEN assert cases are returned
    assert cases

    # THEN assert correct case was returned
    assert cases[0].internal_id == case_id


def test_get_cases_to_process_when_no_case(
    case_id_does_not_exist: str,
    caplog,
    cli_runner: CliRunner,
    helpers: StoreHelpers,
    populated_compress_context: CGConfig,
):
    """Test get cases to proces when there are no cases to compress."""
    caplog.set_level(logging.DEBUG)
    status_db: Store = populated_compress_context.status_db

    # WHEN running the compress command
    cases: List[Family] = get_cases_to_process(
        case_id=case_id_does_not_exist, days_back=1, store=status_db
    )

    # THEN assert no cases where found
    assert not cases

    # THEN assert we log no cases where found
    assert f"Could not find case {case_id_does_not_exist}" in caplog.text


def test_compress_fastq_cli_no_family(compress_context: CGConfig, cli_runner: CliRunner, caplog):
    """Test to run the compress command with a database without samples,"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, [], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0

    # THEN assert it was communicated that no families where found
    assert "No cases to compress" in caplog.text


def test_compress_fastq_cli_case_id_no_family(
    case_id_does_not_exist: str, compress_context: CGConfig, cli_runner: CliRunner, caplog
):
    """Test to run the compress command when no families are found."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--case-id", case_id_does_not_exist], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0

    # THEN assert it was communicated that no families where found
    assert f"Could not find case {case_id_does_not_exist}" in caplog.text


def test_compress_fastq_cli_case_id(
    case_id: str,
    caplog,
    cli_runner: CliRunner,
    helpers: StoreHelpers,
    mocker,
    populated_compress_context: CGConfig,
):
    """Test to run the compress command with a specified case id."""
    caplog.set_level(logging.DEBUG)
    status_db: Store = populated_compress_context.status_db

    # GIVEN a context with a case that can be compressed

    valid_compressable_case: Family = helpers.add_case(
        store=status_db,
        name=case_id,
        internal_id=case_id,
        data_analysis=Pipeline.MIP_DNA,
        action=None,
    )
    valid_compressable_case.created_at = dt.datetime.now() - dt.timedelta(days=1000)
    sample1 = helpers.add_sample(store=status_db, internal_id="ACCR9000")
    sample2 = helpers.add_sample(store=status_db, internal_id="ACCR9001")
    helpers.add_relationship(
        store=status_db,
        sample=sample1,
        case=valid_compressable_case,
    )
    helpers.add_relationship(
        store=status_db,
        sample=sample2,
        case=valid_compressable_case,
    )
    status_db.session.commit()

    # GIVEN no adjusting according to readsa
    mocker.patch(MOCK_SET_MEM_ACCORDING_TO_READS_PATH, return_value=None)

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--case-id", case_id], obj=populated_compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0

    # THEN assert it was communicated that no families where found
    assert f"individuals in 1 (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_multiple_family(
    caplog, cli_runner: CliRunner, mocker, populated_multiple_compress_context: CGConfig
):
    """Test to run the compress command with multiple families."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a database with multiple families
    nr_cases = populated_multiple_compress_context.status_db._get_query(table=Family).count()
    assert nr_cases > 1

    # GIVEN no adjusting according to readsa
    mocker.patch(MOCK_SET_MEM_ACCORDING_TO_READS_PATH, return_value=None)

    # WHEN running the compress command
    res = cli_runner.invoke(
        fastq_cmd, ["--number-of-conversions", nr_cases], obj=populated_multiple_compress_context
    )

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"individuals in {nr_cases} (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_multiple_set_limit(
    caplog, cli_runner: CliRunner, mocker, populated_multiple_compress_context: CGConfig
):
    """Test to run the compress command with multiple families and use a limit."""
    compress_context = populated_multiple_compress_context
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with more families than the limit
    nr_cases = compress_context.status_db._get_query(table=Family).count()
    limit = 5
    assert nr_cases > limit

    # GIVEN no adjusting according to readsa
    mocker.patch(MOCK_SET_MEM_ACCORDING_TO_READS_PATH, return_value=None)

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--number-of-conversions", limit], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated no more than the limited number of cases was compressed
    assert f"individuals in {limit} (completed) cases where compressed" in caplog.text



from typing import Dict, List, Optional, Union

from cg.constants import Pipeline
from cg.constants.gene_panel import GENOME_BUILD_38
from cg.constants.pedigree import Pedigree
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Family
from cg.utils import Process


class MipRNAAnalysisAPI(MipAnalysisAPI):
    def __init__(self, config: CGConfig, pipeline: Pipeline = Pipeline.MIP_RNA):
        super().__init__(config, pipeline)

    @property
    def root(self) -> str:
        return self.config.mip_rd_rna.root

    @property
    def conda_binary(self) -> str:
        return self.config.mip_rd_rna.conda_binary

    @property
    def conda_env(self) -> str:
        return self.config.mip_rd_rna.conda_env

    @property
    def mip_pipeline(self) -> str:
        return self.config.mip_rd_rna.pipeline

    @property
    def script(self) -> str:
        return self.config.mip_rd_rna.script

    @property
    def threshold_reads(self):
        return True

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=f"{self.script} {self.mip_pipeline}",
                conda_binary=f"{self.conda_binary}" if self.conda_binary else None,
                config=self.config.mip_rd_rna.mip_config,
                environment=self.conda_env,
            )
        return self._process

    def config_sample(
        self, link_obj, panel_bed: Optional[str] = None
    ) -> Dict[str, Union[str, int]]:
        sample_data: Dict[str, Union[str, int]] = self.get_sample_data(link_obj)
        if link_obj.mother:
            sample_data[Pedigree.MOTHER.value]: str = link_obj.mother.internal_id
        if link_obj.father:
            sample_data[Pedigree.FATHER.value]: str = link_obj.father.internal_id
        return sample_data

    def panel(self, case_id: str, genome_build: str = GENOME_BUILD_38) -> List[str]:
        """Create the aggregated gene panel file"""
        case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        all_panels = self.convert_panels(case_obj.customer.internal_id, case_obj.panels)
        return self.scout_api.export_panels(build=genome_build, panels=all_panels)



from enum import Enum
from typing import Callable, List, Optional

from sqlalchemy.orm import Query

from cg.store.models import Customer, Pool


def filter_pools_by_customer_id(pools: Query, customer_ids: List[int], **kwargs) -> Query:
    """Return pools by customer id."""
    return pools.filter(Pool.customer_id.in_(customer_ids))


def filter_pools_by_name_enquiry(pools: Query, name_enquiry: str, **kwargs) -> Query:
    """Return pools by name enquiry."""
    return pools.filter(Pool.name.like(f"%{name_enquiry}%"))


def filter_pools_by_order_enquiry(pools: Query, order_enquiry: str, **kwargs) -> Query:
    """Return pools by order enquiry."""
    return pools.filter(Pool.order.like(f"%{order_enquiry}%"))


def filter_pools_by_entry_id(pools: Query, entry_id: int, **kwargs) -> Query:
    """Return pools by entry id."""
    return pools.filter(Pool.id == entry_id)


def filter_pools_by_name(pools: Query, name: str, **kwargs) -> Query:
    """Return pools by name."""
    return pools.filter(Pool.name == name)


def filter_pools_is_received(pools: Query, **kwargs) -> Query:
    """Return received pools."""
    return pools.filter(Pool.received_at.isnot(None))


def filter_pools_is_not_received(pools: Query, **kwargs) -> Query:
    """Return not received pools."""
    return pools.filter(Pool.received_at.is_(None))


def filter_pools_is_delivered(pools: Query, **kwargs) -> Query:
    """Return delivered pools."""
    return pools.filter(Pool.delivered_at.isnot(None))


def filter_pools_is_not_delivered(pools: Query, **kwargs) -> Query:
    """Return not delivered pools."""
    return pools.filter(Pool.delivered_at.is_(None))


def filter_pools_by_invoice_id(pools: Query, invoice_id: int, **kwargs):
    """Return a pool by an invoice id."""
    return pools.filter(Pool.invoice_id == invoice_id)


def filter_pools_without_invoice_id(pools: Query, **kwargs) -> Query:
    """Return pools without an invoice id."""
    return pools.filter(Pool.invoice_id.is_(None))


def filter_pools_do_invoice(pools: Query, **kwargs) -> Query:
    """Return pools marked for invoicing."""
    return pools.filter(Pool.no_invoice.is_(False))


def filter_pools_by_customer(pools: Query, customer: Customer, **kwargs) -> Query:
    """Return pools by customer id."""
    return pools.filter(Pool.customer == customer)


def apply_pool_filter(
    filter_functions: List[Callable],
    pools: Query,
    invoice_id: Optional[int] = None,
    entry_id: Optional[int] = None,
    name: Optional[str] = None,
    customer_ids: Optional[List[int]] = None,
    name_enquiry: Optional[str] = None,
    order_enquiry: Optional[str] = None,
    customer: Optional[Customer] = None,
) -> Query:
    """Apply filtering functions to the pool queries and return filtered results"""

    for filter_function in filter_functions:
        pools: Query = filter_function(
            pools=pools,
            invoice_id=invoice_id,
            entry_id=entry_id,
            name=name,
            customer_ids=customer_ids,
            name_enquiry=name_enquiry,
            order_enquiry=order_enquiry,
            customer=customer,
        )
    return pools


class PoolFilter(Enum):
    """Define Pool filter functions."""

    FILTER_BY_ENTRY_ID: Callable = filter_pools_by_entry_id
    FILTER_BY_NAME: Callable = filter_pools_by_name
    FILTER_IS_RECEIVED: Callable = filter_pools_is_received
    FILTER_IS_NOT_RECEIVED: Callable = filter_pools_is_not_received
    FILTER_IS_DELIVERED: Callable = filter_pools_is_delivered
    FILTER_IS_NOT_DELIVERED: Callable = filter_pools_is_not_delivered
    FILTER_BY_INVOICE_ID: Callable = filter_pools_by_invoice_id
    FILTER_WITHOUT_INVOICE_ID: Callable = filter_pools_without_invoice_id
    FILTER_DO_INVOICE: Callable = filter_pools_do_invoice
    FILTER_BY_CUSTOMER_ID: Callable = filter_pools_by_customer_id
    FILTER_BY_NAME_ENQUIRY: Callable = filter_pools_by_name_enquiry
    FILTER_BY_ORDER_ENQUIRY: Callable = filter_pools_by_order_enquiry
    FILTER_BY_CUSTOMER: Callable = filter_pools_by_customer



"""Module for Balsamic PON Analysis API."""

import logging
from pathlib import Path
from typing import List

from cg.constants.constants import Pipeline
from cg.constants.indexes import ListIndexes
from cg.exc import BalsamicStartError
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Family
from cg.utils.utils import build_command_from_dict

LOG = logging.getLogger(__name__)


class BalsamicPonAnalysisAPI(BalsamicAnalysisAPI):
    """Handles communication between Balsamic PON processes and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.BALSAMIC_PON,
    ):
        super().__init__(config=config, pipeline=pipeline)

    def config_case(
        self,
        case_id: str,
        gender: str,
        genome_version: str,
        panel_bed: str,
        pon_cnn: str,
        observations: List[str],
        dry_run: bool = False,
    ) -> None:
        """Creates a config file for BALSAMIC PON analysis."""
        case: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_parameters: dict = self.get_sample_params(case_id=case_id, panel_bed=panel_bed)
        if not sample_parameters:
            LOG.error(f"{case_id} has no samples tagged for Balsamic PON analysis")
            raise BalsamicStartError()
        verified_panel_bed: str = self.get_verified_bed(
            panel_bed=panel_bed, sample_data=sample_parameters
        )
        options: List[str] = build_command_from_dict(
            {
                "--case-id": case.internal_id,
                "--analysis-dir": self.root_dir,
                "--fastq-path": self.get_sample_fastq_destination_dir(case),
                "--panel-bed": verified_panel_bed,
                "--genome-version": genome_version,
                "--balsamic-cache": self.balsamic_cache,
                "--version": self.get_next_pon_version(verified_panel_bed),
            }
        )
        parameters: List[str] = ["config", "pon"] + options
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def get_case_config_path(self, case_id: str) -> Path:
        """Returns the BALSAMIC PON config path."""
        return Path(self.root_dir, case_id, case_id + "_PON.json")

    def get_next_pon_version(self, panel_bed: str) -> str:
        """Returns the next PON version to be generated."""
        latest_pon_file: str = self.get_latest_pon_file(panel_bed=panel_bed)
        next_version = (
            int(Path(latest_pon_file).stem.split("_v")[ListIndexes.LAST.value]) + 1
            if latest_pon_file
            else 1
        )
        return "v" + str(next_version)



"""Tests for the creation of the sample sheet"""
from typing import List

from cg.apps.demultiplex.sample_sheet.create import get_sample_sheet_creator
from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreator,
    SampleSheetCreatorBcl2Fastq,
    SampleSheetCreatorBCLConvert,
)
from cg.constants.demultiplexing import BclConverter
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData


def test_sample_sheet_creator_factory_novaseq_6000(
    novaseq_flow_cell_demultiplexed_with_bcl2fastq: FlowCellDirectoryData,
    lims_novaseq_bcl2fastq_samples: List[FlowCellSampleBcl2Fastq],
):
    """Test that a sample sheet creator defined with NovaSeq6000 data is V1."""
    # GIVEN a NovaSeq6000 flow cell and a list of NovaSeq6000 samples

    # GIVEN that the flow cell is demultiplexed with bcl2fastq
    assert novaseq_flow_cell_demultiplexed_with_bcl2fastq.bcl_converter == BclConverter.BCL2FASTQ

    # WHEN defining the sample sheet creator
    sample_sheet_creator: SampleSheetCreator = get_sample_sheet_creator(
        flow_cell=novaseq_flow_cell_demultiplexed_with_bcl2fastq,
        lims_samples=lims_novaseq_bcl2fastq_samples,
        force=False,
    )

    # THEN no error is raised and the sample sheet creator is a Bcl2Fastq sample sheet creator
    assert isinstance(sample_sheet_creator, SampleSheetCreatorBcl2Fastq)


def test_sample_sheet_creator_factory_BCL_convert(
    novaseq_x_flow_cell: FlowCellDirectoryData,
    lims_novaseq_bcl_convert_samples: List[FlowCellSampleBCLConvert],
):
    """Test that a sample sheet creator defined with BCL convert data is BCL Convert."""
    # GIVEN a NovaSeqX flow cell and a list of NovaSeqX samples that is demultiplexed wit BCL Convert
    assert novaseq_x_flow_cell.bcl_converter == BclConverter.DRAGEN

    # WHEN defining the sample sheet creator
    sample_sheet_creator: SampleSheetCreator = get_sample_sheet_creator(
        flow_cell=novaseq_x_flow_cell,
        lims_samples=lims_novaseq_bcl_convert_samples,
        force=False,
    )

    # THEN the sample sheet creator is a BCL Convert sample sheet creator
    assert isinstance(sample_sheet_creator, SampleSheetCreatorBCLConvert)



"""All models aggregated in a base class"""
from dataclasses import dataclass
from typing import Callable, List, Optional, Type

from sqlalchemy import and_, func
from sqlalchemy.orm import Query, Session

from cg.store.filters.status_case_filters import CaseFilter, apply_case_filter
from cg.store.filters.status_customer_filters import (
    CustomerFilter,
    apply_customer_filter,
)
from cg.store.filters.status_sample_filters import SampleFilter, apply_sample_filter
from cg.store.models import (
    Analysis,
    Application,
    ApplicationVersion,
    Customer,
    Family,
    FamilySample,
    Flowcell,
)
from cg.store.models import Model as ModelBase
from cg.store.models import Sample
from cg.utils.date import get_date_days_ago


@dataclass
class BaseHandler:
    """All models in one base class."""

    def __init__(self, session: Session):
        self.session = session

    def _get_query(self, table: Type[ModelBase]) -> Query:
        """Return a query for the given table."""
        return self.session.query(table)

    def _get_outer_join_cases_with_analyses_query(self) -> Query:
        """Return a query for all cases in the database with an analysis."""
        return (
            self._get_query(table=Family)
            .outerjoin(Analysis)
            .join(
                Family.links,
                FamilySample.sample,
                ApplicationVersion,
                Application,
            )
        )

    def _get_join_cases_with_samples_query(self) -> Query:
        """Return a join query for all cases in the database with samples."""
        return self._get_query(table=Family).join(
            Family.links, FamilySample.sample, Family.customer
        )

    def _get_join_analysis_case_query(self) -> Query:
        """Return join analysis case query."""
        return self._get_query(table=Analysis).join(Analysis.family)

    def _get_join_case_sample_query(self) -> Query:
        """Return join case sample query."""
        return self._get_query(table=FamilySample).join(FamilySample.family, FamilySample.sample)

    def _get_join_case_and_sample_query(self) -> Query:
        """Return join case sample query."""
        return self._get_query(table=Family).join(Family.links, FamilySample.sample)

    def _get_join_sample_and_customer_query(self) -> Query:
        """Return join sample and customer query."""
        return self._get_query(table=Sample).join(Customer)

    def _get_join_flow_cell_sample_links_query(self) -> Query:
        """Return join flow cell samples and relationship query."""
        return self._get_query(table=Flowcell).join(Flowcell.samples, Sample.links)

    def _get_join_sample_family_query(self) -> Query:
        """Return a join sample case relationship query."""
        return self._get_query(table=Sample).join(Family.links, FamilySample.sample)

    def _get_join_sample_application_version_query(self) -> Query:
        """Return join sample to application version query."""
        return self._get_query(table=Sample).join(
            Sample.application_version, ApplicationVersion.application
        )

    def _get_join_analysis_sample_family_query(self) -> Query:
        """Return join analysis to sample to case query."""
        return self._get_query(table=Analysis).join(Family, Family.links, FamilySample.sample)

    def _get_subquery_with_latest_case_analysis_date(self) -> Query:
        """Return a subquery with the case internal id and the date of its latest analysis."""
        case_and_date: Query = (
            self._get_join_analysis_case_query()
            .group_by(Family.id)
            .with_entities(Analysis.family_id, func.max(Analysis.started_at).label("started_at"))
            .subquery()
        )
        return case_and_date

    def _get_latest_analyses_for_cases_query(self) -> Query:
        """Return a join query for the latest analysis for each case."""
        analyses: Query = self._get_query(table=Analysis)
        case_and_date_subquery: Query = self._get_subquery_with_latest_case_analysis_date()
        return analyses.join(
            case_and_date_subquery,
            and_(
                Analysis.family_id == case_and_date_subquery.c.family_id,
                Analysis.started_at == case_and_date_subquery.c.started_at,
            ),
        )

    def _get_filtered_case_query(
        self,
        case_action: Optional[str],
        customer_id: str,
        data_analysis: str,
        days: int,
        exclude_customer_id: bool,
        internal_id: str,
        name: str,
        priority: str,
        sample_id: str,
    ) -> Query:
        cases_query: Query = self._get_query(table=Family)
        filter_functions: List[Callable] = []

        filter_case_order_date = None
        if days != 0:
            filter_case_order_date = get_date_days_ago(days_ago=days)
            filter_functions.append(CaseFilter.FILTER_NEW_BY_ORDER_DATE)
        if case_action:
            filter_functions.append(CaseFilter.FILTER_BY_ACTION)
        if priority:
            filter_functions.append(CaseFilter.FILTER_BY_PRIORITY)
        if internal_id:
            filter_functions.append(CaseFilter.FILTER_BY_INTERNAL_ID_SEARCH)
        if name:
            filter_functions.append(CaseFilter.FILTER_BY_NAME_SEARCH)
        if data_analysis:
            filter_functions.append(CaseFilter.FILTER_BY_PIPELINE_SEARCH)

        cases_query = apply_case_filter(
            cases=cases_query,
            filter_functions=filter_functions,
            order_date=filter_case_order_date,
            action=case_action,
            priority=priority,
            internal_id_search=internal_id,
            name_search=name,
            pipeline_search=data_analysis,
        )

        # customer filters
        customer_filters: List[Callable] = []
        if customer_id or exclude_customer_id:
            cases_query = cases_query.join(Family.customer)

        if customer_id:
            customer_filters.append(CustomerFilter.FILTER_BY_INTERNAL_ID)

        if exclude_customer_id:
            customer_filters.append(CustomerFilter.EXCLUDE_INTERNAL_ID)

        cases_query = apply_customer_filter(
            customers=cases_query,
            filter_functions=customer_filters,
            customer_internal_id=customer_id,
            exclude_customer_internal_id=exclude_customer_id,
        )

        # sample filters
        if sample_id:
            cases_query = cases_query.join(Family.links, FamilySample.sample)
            cases_query = apply_sample_filter(
                samples=cases_query,
                filter_functions=[SampleFilter.FILTER_BY_INTERNAL_ID_PATTERN],
                internal_id_pattern=sample_id,
            )
        else:
            cases_query = cases_query.outerjoin(Family.links, FamilySample.sample)

        # other joins
        cases_query = cases_query.outerjoin(Family.analyses, Sample.invoice, Sample.flowcells)
        return cases_query



from pathlib import Path

import pytest
from pydantic.v1 import ValidationError as PydanticValidationError

from cg.constants.constants import FileFormat
from cg.exc import ValidationError
from cg.models.nf_analysis import FileDeliverable


def test_file_deliverables(any_string: str, filled_file: Path):
    """Tests that file deliverable model is successfully created."""
    # GIVEN valid deliverables fields

    # WHEN instantiating a deliverables object
    file_deliverable = FileDeliverable(
        format=FileFormat.TSV,
        id=any_string,
        path=filled_file,
        path_index=filled_file,
        step=any_string,
        tag=any_string,
    )

    # THEN assert that it was successfully created
    assert isinstance(file_deliverable, FileDeliverable)

    # THEN assert that paths are returned as strings
    assert isinstance(file_deliverable.path, str)
    assert isinstance(file_deliverable.path_index, str)


def test_file_deliverables_missing_optional(any_string: str, filled_file: Path):
    """Tests file delivery when an optional field is missing."""
    # GIVEN valid deliverables fields

    # WHEN instantiating a deliverables object
    file_deliverable = FileDeliverable(
        format=FileFormat.TSV,
        id=any_string,
        path=filled_file,
        step=any_string,
        tag=any_string,
    )

    # THEN assert that it was successfully created
    assert isinstance(file_deliverable, FileDeliverable)


def test_file_deliverables_missing_mandatory(
    any_string: str,
    filled_file: Path,
):
    """Tests file delivery when a mandatory field is missing."""
    # WHEN instantiating a deliverables object with a missing field

    # THEN assert that an error is raised
    with pytest.raises(PydanticValidationError) as error:
        FileDeliverable(
            id=any_string,
            path=filled_file,
            step=any_string,
            tag=any_string,
        )
    # THEN assert the error message
    assert "field required (type=value_error.missing)" in str(error.value)


def test_file_deliverables_non_existing_attribute(any_string: str, filled_file: Path):
    """Tests file delivery when a non existing attribute is given."""
    # WHEN instantiating a deliverables object with additional attributes not present in the model
    file_deliverable = FileDeliverable(
        format=FileFormat.TSV,
        id=any_string,
        path=filled_file,
        step=any_string,
        tag=any_string,
        nonexisting=FileFormat.TSV,
    )

    # THEN assert that it was successfully created
    assert isinstance(file_deliverable, FileDeliverable)

    # THEN assert that not existing attribute is not included
    assert "nonexisting" not in file_deliverable.__annotations__.keys()


def test_file_deliverables_non_existing_file(
    any_string: str,
    non_existing_file_path: Path,
):
    """Tests file delivery when a mandatory file does not exist."""
    # WHEN instantiating a deliverables object

    # THEN assert that an error is raised
    with pytest.raises(ValidationError) as error:
        FileDeliverable(
            format=FileFormat.TSV,
            id=any_string,
            path=non_existing_file_path,
            step=any_string,
            tag=any_string,
        )
    # THEN assert the error message
    assert f"Path {str(non_existing_file_path)} does not exist" in str(error.value)



from typing import Dict, List, Tuple

from cg.store.models import SampleLaneSequencingMetrics


def combine_mapped_metrics_with_undetermined(
    mapped_metrics: List[SampleLaneSequencingMetrics],
    undetermined_metrics: List[SampleLaneSequencingMetrics],
) -> List[SampleLaneSequencingMetrics]:
    """Combine metrics for mapped and undetermined reads."""

    metrics: Dict[Tuple[str, int], SampleLaneSequencingMetrics] = {
        (metric.sample_internal_id, metric.flow_cell_lane_number): metric
        for metric in mapped_metrics
    }
    for undetermined_metric in undetermined_metrics:
        key = (undetermined_metric.sample_internal_id, undetermined_metric.flow_cell_lane_number)
        existing_metric: SampleLaneSequencingMetrics = metrics.get(key)

        if existing_metric:
            combined_metric: SampleLaneSequencingMetrics = combine_metrics(
                existing_metric=existing_metric, new_metric=undetermined_metric
            )
            metrics[key] = combined_metric
        else:
            metrics[key] = undetermined_metric
    return list(metrics.values())


def combine_metrics(
    existing_metric: SampleLaneSequencingMetrics, new_metric: SampleLaneSequencingMetrics
) -> SampleLaneSequencingMetrics:
    """Update an existing metric with data from a new metric."""

    combined_q30_percentage: float = weighted_average(
        total_1=existing_metric.sample_total_reads_in_lane,
        percentage_1=existing_metric.sample_base_percentage_passing_q30,
        total_2=new_metric.sample_total_reads_in_lane,
        percentage_2=new_metric.sample_base_percentage_passing_q30,
    )
    combined_mean_quality_score: float = weighted_average(
        total_1=existing_metric.sample_total_reads_in_lane,
        percentage_1=existing_metric.sample_base_mean_quality_score,
        total_2=new_metric.sample_total_reads_in_lane,
        percentage_2=new_metric.sample_base_mean_quality_score,
    )
    combined_reads: int = (
        existing_metric.sample_total_reads_in_lane + new_metric.sample_total_reads_in_lane
    )

    existing_metric.sample_base_percentage_passing_q30 = combined_q30_percentage
    existing_metric.sample_base_mean_quality_score = combined_mean_quality_score
    existing_metric.sample_total_reads_in_lane = combined_reads

    return existing_metric


def weighted_average(total_1: int, percentage_1: float, total_2: int, percentage_2: float) -> float:
    """Calculate the weighted average of two percentages."""
    if total_1 == 0 and total_2 == 0:
        return 0
    return (total_1 * percentage_1 + total_2 * percentage_2) / (total_1 + total_2)



import datetime as dt
from pathlib import Path
from typing import Optional

from dateutil.parser import parse as parse_datestr


def parse_str_to_datetime(value: str) -> Optional[dt.datetime]:
    if value:
        return parse_datestr(value)


def parse_str_to_path(value: str) -> Optional[Path]:
    if value:
        return Path(value)



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, BeforeValidator, ConfigDict, constr
from typing_extensions import Annotated

from cg.constants import DataDelivery, Pipeline
from cg.models.orders.validators.sample_base_validators import snake_case
from cg.store.models import Application, Customer, Family, Pool, Sample


class ControlEnum(str, Enum):
    not_control = ""
    positive = "positive"
    negative = "negative"


class SexEnum(str, Enum):
    male = "male"
    female = "female"
    unknown = "unknown"


class PriorityEnum(str, Enum):
    research = "research"
    standard = "standard"
    priority = "priority"
    express = "express"
    clinical_trials = "clinical_trials"


class ContainerEnum(str, Enum):
    no_container = "No container"
    plate = "96 well plate"
    tube = "Tube"


class StatusEnum(str, Enum):
    affected = "affected"
    unaffected = "unaffected"
    unknown = "unknown"


NAME_PATTERN = r"^[A-Za-z0-9-]*$"


class OrderSample(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)
    age_at_sampling: Annotated[
        Optional[str], BeforeValidator(lambda x: str(x) if x else None)
    ] = None
    application: constr(max_length=Application.tag.property.columns[0].type.length)
    capture_kit: Optional[str] = None
    collection_date: Optional[str] = None
    comment: Optional[constr(max_length=Sample.comment.property.columns[0].type.length)] = None
    concentration: Optional[float] = None
    concentration_sample: Optional[float] = None
    container: Optional[ContainerEnum] = None
    container_name: Optional[str] = None
    control: Optional[str] = None
    customer: Optional[
        constr(max_length=Customer.internal_id.property.columns[0].type.length)
    ] = None
    custom_index: Optional[str] = None
    data_analysis: Pipeline
    data_delivery: DataDelivery
    elution_buffer: Optional[str] = None
    extraction_method: Optional[str] = None
    family_name: Optional[
        constr(
            pattern=NAME_PATTERN,
            min_length=2,
            max_length=Family.name.property.columns[0].type.length,
        )
    ] = None
    father: Optional[
        constr(pattern=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length)
    ] = None
    formalin_fixation_time: Optional[int] = None
    index: Optional[str] = None
    index_number: Optional[str] = None
    index_sequence: Optional[str] = None
    internal_id: Optional[
        constr(max_length=Sample.internal_id.property.columns[0].type.length)
    ] = None
    lab_code: Optional[str] = None
    mother: Optional[
        constr(pattern=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length)
    ] = None
    name: constr(
        pattern=NAME_PATTERN,
        min_length=2,
        max_length=Sample.name.property.columns[0].type.length,
    )
    organism: Optional[str] = None
    organism_other: Optional[str] = None
    original_lab: Optional[str] = None
    original_lab_address: Optional[str] = None
    phenotype_groups: Optional[List[str]] = None
    phenotype_terms: Optional[List[str]] = None
    pool: Optional[constr(max_length=Pool.name.property.columns[0].type.length)] = None
    post_formalin_fixation_time: Optional[int] = None
    pre_processing_method: Optional[str] = None
    priority: Annotated[PriorityEnum, BeforeValidator(snake_case)] = PriorityEnum.standard
    primer: Optional[str] = None
    quantity: Optional[int] = None
    reagent_label: Optional[str] = None
    reference_genome: Optional[
        constr(max_length=Sample.reference_genome.property.columns[0].type.length)
    ] = None
    region: Optional[str] = None
    region_code: Optional[str] = None
    require_qc_ok: bool = False
    rml_plate_name: Optional[str] = None
    selection_criteria: Optional[str] = None
    sex: SexEnum = SexEnum.unknown
    source: Optional[str] = None
    status: StatusEnum = StatusEnum.unknown
    subject_id: Optional[
        constr(pattern=NAME_PATTERN, max_length=Sample.subject_id.property.columns[0].type.length)
    ] = None
    tissue_block_size: Optional[str] = None
    tumour: bool = False
    tumour_purity: Optional[int] = None
    verified_organism: Optional[bool] = None
    volume: Annotated[Optional[str], BeforeValidator(lambda x: str(x))] = None
    well_position: Optional[str] = None
    well_position_rml: Optional[str] = None



import logging
import os
import shutil
from pathlib import Path
from typing import List, Optional

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles

LOG = logging.getLogger(__name__)


def is_demultiplexing_copied(analysis_directory: Path) -> bool:
    """Determine whether the demultiplexing has been copied for the latest analysis for a Novaseqx flow cell."""
    return Path(analysis_directory, DemultiplexingDirsAndFiles.COPY_COMPLETE).exists()


def is_flow_cell_demultiplexed(analysis_directory: Path) -> bool:
    """Determine whether the flow cell has been demultiplexed for the latest analysis for a Novaseqx flow cell."""
    return Path(
        analysis_directory,
        DemultiplexingDirsAndFiles.DATA,
        DemultiplexingDirsAndFiles.ANALYSIS_COMPLETED,
    ).exists()


def is_flow_cell_in_demultiplexed_runs(flow_cell_name: str, demultiplexed_runs: Path) -> bool:
    """Determine whether the flow cell is in the demultiplexed runs directory."""
    return Path(demultiplexed_runs, flow_cell_name).exists()


def get_latest_analysis_path(flow_cell_dir: Path) -> Optional[Path]:
    """
    Get the latest analysis directory for a Novaseqx flow cell.
    The latest analysis directory is the one with the highest integer name.
    """
    analysis_path: Path = Path(flow_cell_dir, DemultiplexingDirsAndFiles.ANALYSIS)
    if not analysis_path.exists():
        return None
    analysis_versions: List[Path] = get_sorted_analysis_versions(analysis_path)
    return analysis_versions[0] if analysis_versions else None


def get_sorted_analysis_versions(analysis_path: Path) -> List[Path]:
    """Get a sorted list of analysis version paths for a Novaseqx flow cell."""

    def sort_by_name(version: Path) -> int:
        return int(version.name)

    analysis_versions_paths: List[Path] = [
        version_path for version_path in analysis_path.iterdir() if version_path.is_dir()
    ]
    return sorted(analysis_versions_paths, key=sort_by_name, reverse=True)


def is_queued_for_post_processing(flow_cell_dir: Path) -> bool:
    """Determine whether the flow cell is queued for post processing."""
    return Path(flow_cell_dir, DemultiplexingDirsAndFiles.QUEUED_FOR_POST_PROCESSING).exists()


def hardlink_flow_cell_analysis_data(flow_cell_dir: Path, demultiplexed_runs_dir: Path) -> None:
    """Create hardlinks to the latest version of the analysis data for a Novaseqx flow cell."""
    analysis_path: Path = get_latest_analysis_path(flow_cell_dir)
    analysis_data_path: Path = Path(analysis_path, DemultiplexingDirsAndFiles.DATA)
    demultiplexed_runs_dir_flow_cell_dir = Path(demultiplexed_runs_dir, flow_cell_dir.name)
    hardlink_tree(src=analysis_data_path, dst=demultiplexed_runs_dir_flow_cell_dir)


def mark_as_demultiplexed(flow_cell_dir: Path) -> None:
    """Create the demux_complete.txt file in the specified flow cell directory."""
    Path(flow_cell_dir, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()


def mark_flow_cell_as_queued_for_post_processing(flow_cell_dir: Path) -> None:
    """Create the queued_for_post_processing.txt file in the specified flow cell directory."""
    Path(flow_cell_dir, DemultiplexingDirsAndFiles.QUEUED_FOR_POST_PROCESSING).touch()


def hardlink_tree(src: Path, dst: Path) -> None:
    """Create hardlinks to the specified source directory in the specified destination directory."""
    shutil.copytree(src=src, dst=dst, copy_function=os.link)


def is_ready_for_post_processing(flow_cell_dir: Path, demultiplexed_runs_dir: Path) -> bool:
    """
    Determine whether the flow cell is ready for post processing.
    The flow cell is ready for post processing if:
    - the demultiplexing has been copied
    - the flow cell has been demultiplexed
    - the flow cell is not in the demultiplexed runs directory
    - the flow cell is not queued for post processing
    """
    analysis_path: Path = get_latest_analysis_path(flow_cell_dir)

    if not analysis_path:
        LOG.debug(f"No analysis path found for flow cell {flow_cell_dir.name}.")
        return False

    flow_cell_is_ready: bool = True

    if not is_demultiplexing_copied(analysis_path):
        LOG.debug(f"Demultiplexing has not been copied for flow cell {flow_cell_dir.name}.")
        flow_cell_is_ready = False
    if not is_flow_cell_demultiplexed(analysis_path):
        LOG.debug(f"Flow cell {flow_cell_dir.name} has not been demultiplexed.")
        flow_cell_is_ready = False

    if is_flow_cell_in_demultiplexed_runs(
        flow_cell_name=flow_cell_dir.name, demultiplexed_runs=demultiplexed_runs_dir
    ):
        LOG.debug(f"Flow cell {flow_cell_dir.name} is already in the demultiplexed runs directory.")
        flow_cell_is_ready = False

    if is_queued_for_post_processing(flow_cell_dir):
        LOG.debug(f"Flow cell {flow_cell_dir.name} is already queued for post processing.")
        flow_cell_is_ready = False

    return flow_cell_is_ready



"""Contains base models to be inherited from in other archive software files."""
from abc import abstractmethod
from typing import List

from housekeeper.store.models import File
from pydantic import BaseModel, ConfigDict

from cg.models.cg_config import DataFlowConfig
from cg.store.models import Sample


class FileAndSample(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    file: File
    sample: Sample


class SampleAndDestination(BaseModel):
    """Contains a sample and the path to the sample bundle in Housekeeper where any retrieved files should be stored."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    sample: Sample
    destination: str


class FileTransferData(BaseModel):
    """Base class for classes representing files to be archived."""

    @classmethod
    @abstractmethod
    def create_from_file_and_sample(
        cls, file: File, sample: Sample, is_archiving: bool
    ) -> "FileTransferData":
        """Instantiates the class from a File and Sample object."""
        pass


class ArchiveHandler:
    """Base class for classes handling different archiving programs."""

    @abstractmethod
    def __init__(self, config: DataFlowConfig):
        """Initiates the ArchiveHandler based on the provided configuration."""
        pass

    @abstractmethod
    def archive_files(self, files_and_samples: List[FileAndSample]):
        """Archives all folders provided, to their corresponding destination,
        as given by sources and destination parameter."""
        pass

    @abstractmethod
    def retrieve_samples(self, samples_and_destinations: List[SampleAndDestination]):
        """Retrieves all files for all samples for the given flowcell."""
        pass

    @abstractmethod
    def retrieve_file(self, file_and_sample: FileAndSample):
        """Retrieves the specified archived file."""
        pass

    @abstractmethod
    def convert_into_transfer_data(
        self, files_and_samples: List[FileAndSample], is_archiving: bool = True
    ) -> List[FileTransferData]:
        """Converts the provided files_and_samples into a list of objects formatted for the specific archiving flow."""
        pass



from enum import Enum
from typing import Callable, List, Optional

from sqlalchemy.orm import Query

from cg.store.models import User


def filter_user_by_email(users: Query, email: str, **kwargs) -> Query:
    """Return user by email."""
    return users.filter(User.email == email)


class UserFilter(Enum):
    """Define User filter functions."""

    FILTER_BY_EMAIL: Callable = filter_user_by_email


def apply_user_filter(
    users: Query,
    filter_functions: List[Callable],
    email: Optional[str] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for filter_function in filter_functions:
        users: Query = filter_function(
            users=users,
            email=email,
        )
    return users



"""Code for uploading coverage reports via CLI"""

import click

from cg.meta.upload.coverage import UploadCoverageApi
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Family

from .utils import suggest_cases_to_upload


@click.command("coverage")
@click.option(
    "-r",
    "--re-upload",
    is_flag=True,
    help="re-upload existing analysis",
)
@click.argument("family_id", required=False)
@click.pass_obj
def upload_coverage(context: CGConfig, re_upload, family_id):
    """Upload coverage from an analysis to Chanjo."""

    click.echo(click.style("----------------- COVERAGE --------------------"))

    status_db: Store = context.status_db

    if not family_id:
        suggest_cases_to_upload(status_db=status_db)
        raise click.Abort

    case_obj: Family = status_db.get_case_by_internal_id(internal_id=family_id)
    upload_coverage_api = UploadCoverageApi(
        status_api=status_db,
        hk_api=context.housekeeper_api,
        chanjo_api=context.chanjo_api,
    )
    coverage_data = upload_coverage_api.data(case_obj.analyses[0])
    upload_coverage_api.upload(coverage_data, replace=re_upload)



"""cg module for cleaning databases and files."""
import logging
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional

import click
from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scout_export import ScoutExportCase
from cg.apps.scout.scoutapi import ScoutAPI
from cg.cli.workflow.commands import (
    balsamic_past_run_dirs,
    balsamic_pon_past_run_dirs,
    balsamic_qc_past_run_dirs,
    balsamic_umi_past_run_dirs,
    fluffy_past_run_dirs,
    microsalt_past_run_dirs,
    mip_dna_past_run_dirs,
    mip_rna_past_run_dirs,
    mutant_past_run_dirs,
    rnafusion_past_run_dirs,
    rsync_past_run_dirs,
)
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.constants.constants import DRY_RUN, SKIP_CONFIRMATION, Pipeline
from cg.constants.housekeeper_tags import ALIGNMENT_FILE_TAGS, ScoutTag
from cg.exc import CleanFlowCellFailedError
from cg.meta.clean.api import CleanAPI
from cg.meta.clean.clean_flow_cells import CleanFlowCellAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Analysis
from cg.utils.date import get_date_days_ago, get_timedelta_from_date
from cg.utils.dispatcher import Dispatcher
from cg.utils.files import get_directories_in_path

CHECK_COLOR = {True: "green", False: "red"}
LOG = logging.getLogger(__name__)
FLOW_CELL_OUTPUT_HEADERS = [
    "Flow cell run name",
    "Flow cell id",
    "Correct name?",
    "Exists in statusdb?",
    "Fastq files in HK?",
    "Spring files in HK?",
    "Files on disk?",
    "Check passed?",
]


@click.group()
def clean():
    """Clean up processes."""
    return


for sub_cmd in [
    balsamic_past_run_dirs,
    balsamic_qc_past_run_dirs,
    balsamic_umi_past_run_dirs,
    balsamic_pon_past_run_dirs,
    fluffy_past_run_dirs,
    mip_dna_past_run_dirs,
    mip_rna_past_run_dirs,
    mutant_past_run_dirs,
    rnafusion_past_run_dirs,
    rsync_past_run_dirs,
    microsalt_past_run_dirs,
]:
    clean.add_command(sub_cmd)


@clean.command("hk-alignment-files")
@click.argument("bundle")
@DRY_RUN
@SKIP_CONFIRMATION
@click.pass_obj
def hk_alignment_files(
    context: CGConfig, bundle: str, yes: bool = False, dry_run: bool = False
) -> None:
    """Clean up alignment files in Housekeeper bundle."""
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    for tag in ALIGNMENT_FILE_TAGS:
        tag_files = set(housekeeper_api.get_files(bundle=bundle, tags=[tag]))

        if not tag_files:
            LOG.warning(
                f"Could not find any files ready for cleaning for bundle {bundle} and tag {tag}"
            )

        for hk_file in tag_files:
            if not (yes or click.confirm(_get_confirm_question(bundle, hk_file))):
                continue

            file_path: Path = Path(hk_file.full_path)
            if hk_file.is_included and file_path.exists():
                LOG.info(f"Unlinking {file_path}")
                if not dry_run:
                    file_path.unlink()

            LOG.info(f"Deleting {file_path} from database")
            if not dry_run:
                housekeeper_api.delete_file(file_id=hk_file.id)
                housekeeper_api.commit()


@clean.command("scout-finished-cases")
@click.option(
    "--days-old",
    type=int,
    default=300,
    help="Clean alignment files with analysis dates older then given number of days",
)
@SKIP_CONFIRMATION
@DRY_RUN
@click.pass_context
def scout_finished_cases(
    context: click.Context, days_old: int, yes: bool = False, dry_run: bool = False
) -> None:
    """Clean up of solved and archived Scout cases."""
    scout_api: ScoutAPI = context.obj.scout_api
    bundles: List[str] = []
    for status in [ScoutTag.ARCHIVED.value, ScoutTag.SOLVED.value]:
        cases: List[ScoutExportCase] = scout_api.get_cases(status=status, reruns=False)
        cases_added: int = 0
        for case in cases:
            analysis_time_delta: timedelta = get_timedelta_from_date(date=case.analysis_date)
            if analysis_time_delta.days > days_old:
                bundles.append(case.id)
                cases_added += 1
        LOG.info(f"{cases_added} cases marked for alignment files removal")

    for bundle in bundles:
        context.invoke(hk_alignment_files, bundle=bundle, yes=yes, dry_run=dry_run)


@clean.command("hk-case-bundle-files")
@click.option(
    "--days-old",
    type=int,
    default=365,
    help="Clean all files with analysis dates older then given number of days",
)
@DRY_RUN
@click.pass_context
def hk_case_bundle_files(context: CGConfig, days_old: int, dry_run: bool = False) -> None:
    """Clean up all non-protected files for all pipelines."""
    housekeeper_api: HousekeeperAPI = context.obj.housekeeper_api
    clean_api: CleanAPI = CleanAPI(status_db=context.obj.status_db, housekeeper_api=housekeeper_api)

    size_cleaned: int = 0
    version_file: File
    for version_file in clean_api.get_unprotected_existing_bundle_files(
        before=get_date_days_ago(days_ago=days_old)
    ):
        file_path: Path = Path(version_file.full_path)
        file_size: int = file_path.stat().st_size
        size_cleaned += file_size
        if dry_run:
            LOG.info(f"Dry run: {dry_run}. Keeping file {file_path}")
            continue

        file_path.unlink()
        housekeeper_api.delete_file(file_id=version_file.id)
        housekeeper_api.commit()
        LOG.info(f"Removed file {file_path}. Dry run: {dry_run}")

    LOG.info(f"Process freed {round(size_cleaned * 0.0000000001, 2)} GB. Dry run: {dry_run}")


@clean.command("hk-bundle-files")
@click.option("-c", "--case-id", type=str, required=False)
@click.option("-p", "--pipeline", type=Pipeline, required=False)
@click.option("-t", "--tags", multiple=True, required=True)
@click.option("-o", "--days-old", type=int, default=30)
@DRY_RUN
@click.pass_obj
def hk_bundle_files(
    context: CGConfig,
    case_id: Optional[str],
    tags: list,
    days_old: Optional[int],
    pipeline: Optional[Pipeline],
    dry_run: bool,
):
    """Remove files found in Housekeeper bundles."""

    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db

    date_threshold: datetime = get_date_days_ago(days_ago=days_old)

    function_dispatcher: Dispatcher = Dispatcher(
        functions=[
            status_db.get_analyses_started_at_before,
            status_db.get_analyses_for_case_and_pipeline_started_at_before,
            status_db.get_analyses_for_pipeline_started_at_before,
            status_db.get_analyses_for_case_started_at_before,
        ],
        input_dict={
            "case_internal_id": case_id,
            "pipeline": pipeline,
            "started_at_before": date_threshold,
        },
    )
    analyses: List[Analysis] = function_dispatcher()

    size_cleaned: int = 0
    for analysis in analyses:
        LOG.info(f"Cleaning analysis {analysis}")
        bundle_name: str = analysis.family.internal_id
        hk_bundle_version: Optional[Version] = housekeeper_api.version(
            bundle=bundle_name, date=analysis.started_at
        )
        if not hk_bundle_version:
            LOG.warning(
                f"Version not found for "
                f"bundle:{bundle_name}; "
                f"pipeline: {analysis.pipeline}; "
                f"date {analysis.started_at}"
            )
            continue

        LOG.info(
            f"Version found for "
            f"bundle:{bundle_name}; "
            f"pipeline: {analysis.pipeline}; "
            f"date {analysis.started_at}"
        )
        version_files: List[File] = housekeeper_api.get_files(
            bundle=analysis.family.internal_id, tags=tags, version=hk_bundle_version.id
        ).all()
        for version_file in version_files:
            file_path: Path = Path(version_file.full_path)
            if not file_path.exists():
                LOG.info(f"File {file_path} not on disk.")
                continue
            LOG.info(f"File {file_path} found on disk.")
            file_size = file_path.stat().st_size
            size_cleaned += file_size
            if dry_run:
                continue

            file_path.unlink()
            housekeeper_api.delete_file(version_file.id)
            housekeeper_api.commit()
            LOG.info(f"Removed file {file_path}. Dry run: {dry_run}")

    LOG.info(f"Process freed {round(size_cleaned * 0.0000000001, 2)}GB. Dry run: {dry_run}")


@clean.command("flow-cells")
@DRY_RUN
@click.pass_obj
def clean_flow_cells(context: CGConfig, dry_run: bool):
    """Remove flow cells from the flow cells and demultiplexed runs folder."""

    directories_to_check: List[Path] = []
    for path in [Path(context.flow_cells_dir), Path(context.demultiplexed_flow_cells_dir)]:
        directories_to_check.extend(get_directories_in_path(path))
    exit_code = EXIT_SUCCESS
    for flow_cell_directory in directories_to_check:
        try:
            clean_flow_cell_api = CleanFlowCellAPI(
                flow_cell_path=flow_cell_directory,
                status_db=context.status_db,
                housekeeper_api=context.housekeeper_api,
                dry_run=dry_run,
            )
            clean_flow_cell_api.delete_flow_cell_directory()
        except CleanFlowCellFailedError as error:
            LOG.error(repr(error))
            exit_code = EXIT_FAIL
    if exit_code:
        click.Abort


def _get_confirm_question(bundle, file_obj) -> str:
    """Return confirmation question."""
    return (
        f"{bundle}: remove file from file system and database: {file_obj.full_path}"
        if file_obj.is_included
        else f"{bundle}: remove file from database: {file_obj.full_path}"
    )



import logging
from typing import Optional

from pydantic.v1 import BaseModel, EmailStr, Field
from typing_extensions import Literal

from cg.apps.coverage import ChanjoAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.gens import GensAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.loqus import LoqusdbAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.observations import LoqusdbInstance
from cg.constants.priority import SlurmQos
from cg.store import Store

LOG = logging.getLogger(__name__)


class Sequencers(BaseModel):
    hiseqx: str
    hiseqga: str
    novaseq: str


class SlurmConfig(BaseModel):
    account: str
    hours: Optional[int]
    mail_user: EmailStr
    memory: Optional[int]
    number_tasks: Optional[int]
    conda_env: Optional[str]
    qos: SlurmQos = SlurmQos.LOW


class BackupConfig(BaseModel):
    encrypt_dir: str
    slurm_flow_cell_encryption: SlurmConfig


class HousekeeperConfig(BaseModel):
    database: str
    root: str


class DemultiplexConfig(BaseModel):
    slurm: SlurmConfig


class TrailblazerConfig(BaseModel):
    service_account: str
    service_account_auth_file: str
    host: str


class StatinaConfig(BaseModel):
    host: Optional[str]
    user: str
    key: str
    api_url: str
    upload_path: str
    auth_path: str


class CommonAppConfig(BaseModel):
    binary_path: str
    config_path: Optional[str]


class FluffyUploadConfig(BaseModel):
    user: str
    password: str
    host: str
    remote_path: str
    port: int


class FluffyConfig(CommonAppConfig):
    root_dir: str
    sftp: FluffyUploadConfig


class LimsConfig(BaseModel):
    host: str
    username: str
    password: str


class CrunchyConfig(BaseModel):
    conda_binary: Optional[str] = None
    cram_reference: str
    slurm: SlurmConfig


class MutaccAutoConfig(CommonAppConfig):
    padding: int = 300


class BalsamicConfig(CommonAppConfig):
    balsamic_cache: str
    bed_path: str
    binary_path: str
    conda_env: str
    loqusdb_path: str
    pon_path: str
    root: str
    slurm: SlurmConfig
    swegen_path: str


class MutantConfig(BaseModel):
    binary_path: str
    conda_binary: Optional[str] = None
    conda_env: str
    root: str


class MipConfig(BaseModel):
    conda_binary: Optional[str] = None
    conda_env: str
    mip_config: str
    pipeline: str
    root: str
    script: str


class RnafusionConfig(CommonAppConfig):
    root: str
    references: str
    binary_path: str
    pipeline_path: str
    conda_env: str
    compute_env: str
    profile: str
    conda_binary: Optional[str] = None
    launch_directory: str
    revision: str
    slurm: SlurmConfig
    tower_binary_path: str
    tower_pipeline: str


class TaxprofilerConfig(CommonAppConfig):
    root: str
    binary_path: str
    conda_env: str
    profile: str
    pipeline_path: str
    revision: str
    conda_binary: Optional[str] = None
    hostremoval_reference: str
    databases: str
    slurm: SlurmConfig
    tower_binary_path: str
    tower_pipeline: str


class MicrosaltConfig(BaseModel):
    binary_path: str
    conda_binary: Optional[str] = None
    conda_env: str
    queries_path: str
    root: str


class GisaidConfig(CommonAppConfig):
    submitter: str
    log_dir: str
    logwatch_email: EmailStr
    upload_password: str
    upload_cid: str


class DataDeliveryConfig(BaseModel):
    destination_path: str
    covid_destination_path: str
    covid_source_path = str
    covid_report_path: str
    account: str
    base_path: str
    mail_user: str


class EmailBaseSettings(BaseModel):
    sender_email: EmailStr
    smtp_server: str


class FOHMConfig(BaseModel):
    host: str
    port: int
    key: str
    username: str
    valid_uploader: str
    email_sender: str
    email_recipient: str
    email_host: str


class ExternalConfig(BaseModel):
    hasta: str
    caesar: str


class DataFlowConfig(BaseModel):
    database_name: str
    user: str
    password: str
    url: str
    local_storage: str
    archive_repository: str


class CGConfig(BaseModel):
    database: str
    environment: Literal["production", "stage"] = "stage"
    madeline_exe: str
    delivery_path: str
    max_flowcells: Optional[int]
    email_base_settings: EmailBaseSettings
    flow_cells_dir: str
    demultiplexed_flow_cells_dir: str

    # Base APIs that always should exist
    status_db_: Store = None
    housekeeper: HousekeeperConfig
    housekeeper_api_: HousekeeperAPI = None

    # App APIs that can be instantiated in CGConfig
    backup: BackupConfig = None
    chanjo: CommonAppConfig = None
    chanjo_api_: ChanjoAPI = None
    crunchy: CrunchyConfig = None
    crunchy_api_: CrunchyAPI = None
    data_delivery: DataDeliveryConfig = Field(None, alias="data-delivery")
    data_flow_config: Optional[DataFlowConfig] = None
    demultiplex: DemultiplexConfig = None
    demultiplex_api_: DemultiplexingAPI = None
    encryption: Optional[CommonAppConfig] = None
    external: ExternalConfig = None
    genotype: CommonAppConfig = None
    genotype_api_: GenotypeAPI = None
    gens: CommonAppConfig = None
    gens_api_: GensAPI = None
    hermes: CommonAppConfig = None
    hermes_api_: HermesApi = None
    lims: LimsConfig = None
    lims_api_: LimsAPI = None
    loqusdb: CommonAppConfig = Field(None, alias=LoqusdbInstance.WGS.value)
    loqusdb_api_: LoqusdbAPI = None
    loqusdb_wes: CommonAppConfig = Field(None, alias=LoqusdbInstance.WES.value)
    loqusdb_somatic: CommonAppConfig = Field(None, alias=LoqusdbInstance.SOMATIC.value)
    loqusdb_tumor: CommonAppConfig = Field(None, alias=LoqusdbInstance.TUMOR.value)
    madeline_api_: MadelineAPI = None
    mutacc_auto: MutaccAutoConfig = Field(None, alias="mutacc-auto")
    mutacc_auto_api_: MutaccAutoAPI = None
    pigz: Optional[CommonAppConfig] = None
    pdc: Optional[CommonAppConfig] = None
    scout: CommonAppConfig = None
    scout_api_: ScoutAPI = None
    tar: Optional[CommonAppConfig] = None
    trailblazer: TrailblazerConfig = None
    trailblazer_api_: TrailblazerAPI = None

    # Meta APIs that will use the apps from CGConfig
    balsamic: BalsamicConfig = None
    statina: StatinaConfig = None
    fohm: Optional[FOHMConfig] = None
    fluffy: FluffyConfig = None
    microsalt: MicrosaltConfig = None
    gisaid: GisaidConfig = None
    mip_rd_dna: MipConfig = Field(None, alias="mip-rd-dna")
    mip_rd_rna: MipConfig = Field(None, alias="mip-rd-rna")
    mutant: MutantConfig = None
    rnafusion: RnafusionConfig = Field(None, alias="rnafusion")
    taxprofiler: TaxprofilerConfig = Field(None, alias="taxprofiler")

    # These are meta APIs that gets instantiated in the code
    meta_apis: dict = {}

    class Config:
        arbitrary_types_allowed = True
        fields = {
            "chanjo_api_": "chanjo_api",
            "crunchy_api_": "crunchy_api",
            "demultiplex_api_": "demultiplex_api",
            "genotype_api_": "genotype_api",
            "gens_api_": "gens_api",
            "hermes_api_": "hermes_api",
            "housekeeper_api_": "housekeeper_api",
            "lims_api_": "lims_api",
            "loqusdb_api_": "loqusdb_api",
            "madeline_api_": "madeline_api",
            "mutacc_auto_api_": "mutacc_auto_api",
            "scout_api_": "scout_api",
            "status_db_": "status_db",
            "trailblazer_api_": "trailblazer_api",
        }

    @property
    def chanjo_api(self) -> ChanjoAPI:
        api = self.__dict__.get("chanjo_api_")
        if api is None:
            LOG.debug("Instantiating chanjo api")
            api = ChanjoAPI(config=self.dict())
            self.chanjo_api_ = api
        return api

    @property
    def crunchy_api(self) -> CrunchyAPI:
        api = self.__dict__.get("crunchy_api_")
        if api is None:
            LOG.debug("Instantiating crunchy api")
            api = CrunchyAPI(config=self.dict())
            self.crunchy_api_ = api
        return api

    @property
    def demultiplex_api(self) -> DemultiplexingAPI:
        demultiplex_api = self.__dict__.get("demultiplex_api_")
        if demultiplex_api is None:
            LOG.debug("Instantiating demultiplexing api")
            demultiplex_api = DemultiplexingAPI(
                config=self.dict(), housekeeper_api=self.housekeeper_api
            )
            self.demultiplex_api_ = demultiplex_api
        return demultiplex_api

    @property
    def genotype_api(self) -> GenotypeAPI:
        api = self.__dict__.get("genotype_api_")
        if api is None:
            LOG.debug("Instantiating genotype api")
            api = GenotypeAPI(config=self.dict())
            self.genotype_api_ = api
        return api

    @property
    def gens_api(self) -> GensAPI:
        """Returns Gens API after making sure it has been instantiated."""
        api = self.__dict__.get("gens_api_")
        if api is None:
            LOG.debug("Instantiating gens api")
            api = GensAPI(config=self.dict())
            self.gens_api_ = api
        return api

    @property
    def hermes_api(self) -> HermesApi:
        hermes_api = self.__dict__.get("hermes_api_")
        if hermes_api is None:
            LOG.debug("Instantiating hermes api")
            hermes_api = HermesApi(config=self.dict())
            self.hermes_api_ = hermes_api
        return hermes_api

    @property
    def housekeeper_api(self) -> HousekeeperAPI:
        housekeeper_api = self.__dict__.get("housekeeper_api_")
        if housekeeper_api is None:
            LOG.debug("Instantiating housekeeper api")
            housekeeper_api = HousekeeperAPI(config=self.dict())
            self.housekeeper_api_ = housekeeper_api
        return housekeeper_api

    @property
    def lims_api(self) -> LimsAPI:
        api = self.__dict__.get("lims_api_")
        if api is None:
            LOG.debug("Instantiating lims api")
            api = LimsAPI(config=self.dict())
            self.lims_api_ = api
        return api

    @property
    def loqusdb_api(self) -> LoqusdbAPI:
        api = self.__dict__.get("loqusdb_api_")
        if api is None:
            LOG.debug("Instantiating loqusdb api")
            api: LoqusdbAPI = LoqusdbAPI(
                binary_path=self.loqusdb.binary_path, config_path=self.loqusdb.config_path
            )
            self.loqusdb_api_ = api
        return api

    @property
    def madeline_api(self) -> MadelineAPI:
        api = self.__dict__.get("madeline_api_")
        if api is None:
            LOG.debug("Instantiating madeline api")
            api = MadelineAPI(config=self.dict())
            self.madeline_api_ = api
        return api

    @property
    def mutacc_auto_api(self) -> MutaccAutoAPI:
        api = self.__dict__.get("mutacc_auto_api_")
        if api is None:
            LOG.debug("Instantiating mutacc_auto api")
            api = MutaccAutoAPI(config=self.dict())
            self.mutacc_auto_api_ = api
        return api

    @property
    def scout_api(self) -> ScoutAPI:
        api = self.__dict__.get("scout_api_")
        if api is None:
            LOG.debug("Instantiating scout api")
            api = ScoutAPI(config=self.dict())
            self.scout_api_ = api
        return api

    @property
    def status_db(self) -> Store:
        status_db = self.__dict__.get("status_db_")
        if status_db is None:
            LOG.debug("Instantiating status db")
            status_db = Store(uri=self.database)
            self.status_db_ = status_db
        return status_db

    @property
    def trailblazer_api(self) -> TrailblazerAPI:
        api = self.__dict__.get("trailblazer_api_")
        if api is None:
            LOG.debug("Instantiating trailblazer api")
            api = TrailblazerAPI(config=self.dict())
            self.trailblazer_api_ = api
        return api



from typing import List, Optional

from pydantic import BaseModel

from cg.models.orders.sample_base import OrderSample


# Class for holding information about cases in order
class OrderCase(BaseModel):
    cohorts: Optional[List[str]]
    name: str
    panels: Optional[List[str]]
    priority: str
    samples: List[OrderSample]
    synopsis: Optional[str]


class OrderPool(BaseModel):
    name: str
    data_analysis: str
    data_delivery: Optional[str]
    application: str
    samples: List[OrderSample]


# This is for validating in data
class Orderform(BaseModel):
    comment: Optional[str] = None
    delivery_type: str
    project_type: str
    customer: str
    name: str
    data_analysis: Optional[str] = None
    data_delivery: Optional[str] = None
    ticket: Optional[int] = None
    samples: List[OrderSample]
    cases: Optional[List[OrderCase]] = None



"""RNAfusion related constants."""

RNAFUSION_METRIC_CONDITIONS: dict = {
    "uniquely_mapped_percent": {"norm": "gt", "threshold": 60},
    "PCT_MRNA_BASES": {"norm": "gt", "threshold": 80},
    "PCT_RIBOSOMAL_BASES": {"norm": "lt", "threshold": 5},
    "PERCENT_DUPLICATION": {"norm": "lt", "threshold": 0.7},
}



from pathlib import Path
from typing import List

import pytest


@pytest.fixture
def dsmc_q_archive_output() -> List[str]:
    output: str = """IBM Tivoli Storage Manager
Command Line Backup-Archive Client Interface
  Client Version 7, Release 1, Level 4.0
  Client date/time: 09/22/2023 10:42:42
(c) Copyright by IBM Corporation and other(s) 1990, 2015. All Rights Reserved.

Node Name: HASTA.SCILIFELAB.SE_CLINICAL
Session established with server BLACKHOLE: Linux/x86_64
  Server Version 8, Release 1, Level 9.300
  Server date/time: 09/22/2023 10:42:42  Last access: 09/22/2023 10:42:26

Accessing as node: SLLCLINICAL
             Size  Archive Date - Time    File - Expires on - Description
             ----  -------------------    -------------------------------
           607  B  04/07/2019 07:30:15    /home/hiseq.clinical/ENCRYPT/190329_A00689_0018_AHVKJCDRXX.key.gpg Never Archive Date: 04/07/2019
 1,244,997,334 KB  04/07/2019 04:00:05    /home/hiseq.clinical/ENCRYPT/190329_A00689_0018_AHVKJCDRXX.tar.gz.gpg Never Archive Date: 04/07/2019"""
    return output.splitlines()


@pytest.fixture
def spring_file_path() -> Path:
    """Return spring file path"""
    return Path("/path/to/spring_file.spring")


@pytest.fixture
def backup_file_path() -> str:
    """Return path to a file used in the backup process"""
    return "/path/to/backup_file.extension"


@pytest.fixture
def archived_flow_cell() -> Path:
    """Path of archived flow cell"""
    return Path("/path/to/archived/flow_cell.tar.gz.gpg")


@pytest.fixture
def archived_key() -> Path:
    """Path of archived key"""
    return Path("/path/to/archived/encryption_key.key.gpg")



import logging

from housekeeper.store.models import Version

from cg.apps.lims import LimsAPI
from cg.constants.scout_upload import BALSAMIC_UMI_CASE_TAGS, BALSAMIC_UMI_SAMPLE_TAGS
from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.models.scout.scout_load_config import (
    BalsamicUmiLoadConfig,
    ScoutCancerIndividual,
)
from cg.store.models import Analysis, Sample

LOG = logging.getLogger(__name__)


class BalsamicUmiConfigBuilder(BalsamicConfigBuilder):
    def __init__(self, hk_version_obj: Version, analysis_obj: Analysis, lims_api: LimsAPI):
        super().__init__(
            hk_version_obj=hk_version_obj, analysis_obj=analysis_obj, lims_api=lims_api
        )
        self.case_tags: CaseTags = CaseTags(**BALSAMIC_UMI_CASE_TAGS)
        self.sample_tags: SampleTags = SampleTags(**BALSAMIC_UMI_SAMPLE_TAGS)
        self.load_config: BalsamicUmiLoadConfig = BalsamicUmiLoadConfig(track="cancer")

    def include_sample_files(self, config_sample: ScoutCancerIndividual) -> None:
        LOG.info("Including BALSAMIC specific sample level files")

    def get_balsamic_analysis_type(self, sample: Sample) -> str:
        """Returns a formatted balsamic analysis type"""

        return "panel-umi"



"""Tests for running the demultiplex flowcell command"""
import logging
from pathlib import Path

from click import testing

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.cli.demultiplex.demux import (
    delete_flow_cell,
    demultiplex_all,
    demultiplex_flow_cell,
)
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.meta.demultiplex.housekeeper_storage_functions import (
    add_sample_sheet_path_to_housekeeper,
)
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from tests.meta.demultiplex.conftest import (
    tmp_flow_cell_demux_base_path,
    tmp_flow_cell_run_base_path,
)


def test_demultiplex_flow_cell_dry_run(
    cli_runner: testing.CliRunner,
    tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq: Path,
    demultiplexing_context_for_demux: CGConfig,
    caplog,
):
    caplog.set_level(logging.INFO)

    # GIVEN that all files are present for demultiplexing
    flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(
        tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq
    )
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq,
        flow_cell_name=flow_cell.id,
        hk_api=demultiplexing_context_for_demux.housekeeper_api,
    )

    # GIVEN an out dir that does not exist
    demux_api: DemultiplexingAPI = demultiplexing_context_for_demux.demultiplex_api
    assert demux_api.is_demultiplexing_possible(flow_cell=flow_cell)
    demux_dir: Path = demux_api.flow_cell_out_dir_path(flow_cell)
    unaligned_dir: Path = Path(demux_dir, DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME)
    assert demux_dir.exists() is False
    assert unaligned_dir.exists() is False

    # WHEN starting demultiplexing from the CLI with dry run flag
    result: testing.Result = cli_runner.invoke(
        demultiplex_flow_cell,
        [str(tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq), "--dry-run"],
        obj=demultiplexing_context_for_demux,
    )

    # THEN assert the command exits without problems
    assert result.exit_code == 0

    # THEN assert no results folder was created since it is run in dry run mode
    assert demux_dir.exists() is False
    assert unaligned_dir.exists() is False


def test_demultiplex_bcl2fastq_flow_cell(
    cli_runner: testing.CliRunner,
    tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq: Path,
    demultiplexing_context_for_demux: CGConfig,
    caplog,
    mocker,
):
    caplog.set_level(logging.INFO)

    # GIVEN that all files are present for bcl2fastq demultiplexing
    flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(
        tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq
    )
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq,
        flow_cell_name=flow_cell.id,
        hk_api=demultiplexing_context_for_demux.housekeeper_api,
    )

    # GIVEN an out dir that does not exist
    demux_api: DemultiplexingAPI = demultiplexing_context_for_demux.demultiplex_api
    demux_dir: Path = demux_api.flow_cell_out_dir_path(flow_cell)
    unaligned_dir: Path = Path(demux_dir, DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME)
    assert demux_api.is_demultiplexing_possible(flow_cell=flow_cell)
    assert demux_dir.exists() is False
    assert unaligned_dir.exists() is False
    mocker.patch("cg.apps.tb.TrailblazerAPI.add_pending_analysis")

    # WHEN starting demultiplexing from the CLI with dry run flag
    result: testing.Result = cli_runner.invoke(
        demultiplex_flow_cell,
        [str(tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq)],
        obj=demultiplexing_context_for_demux,
    )

    # THEN assert the command exits sucessfully

    assert result.exit_code == 0

    # THEN assert the results folder was created
    assert demux_dir.exists()
    assert unaligned_dir.exists()

    # THEN assert that the sbatch script was created
    assert demux_api.demultiplex_sbatch_path(flow_cell).exists()


def test_demultiplex_dragen_flowcell(
    cli_runner: testing.CliRunner,
    tmp_flow_cell_directory_bclconvert: Path,
    demultiplexing_context_for_demux: CGConfig,
    tmp_demultiplexed_runs_directory: Path,
    caplog,
    mocker,
):
    caplog.set_level(logging.INFO)

    # GIVEN that all files are present for Dragen demultiplexing

    flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(
        flow_cell_path=tmp_flow_cell_directory_bclconvert, bcl_converter="dragen"
    )
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=tmp_flow_cell_directory_bclconvert,
        flow_cell_name=flow_cell.id,
        hk_api=demultiplexing_context_for_demux.housekeeper_api,
    )

    # GIVEN an out dir that does not exist
    demux_api: DemultiplexingAPI = demultiplexing_context_for_demux.demultiplex_api
    demux_dir: Path = demux_api.flow_cell_out_dir_path(flow_cell)
    assert demux_api.is_demultiplexing_possible(flow_cell=flow_cell)
    assert demux_dir.exists() is False
    mocker.patch("cg.apps.tb.TrailblazerAPI.add_pending_analysis")

    # WHEN starting demultiplexing from the CLI with dry run flag
    result: testing.Result = cli_runner.invoke(
        demultiplex_flow_cell,
        [str(tmp_flow_cell_directory_bclconvert), "-b", "dragen"],
        obj=demultiplexing_context_for_demux,
    )

    # THEN assert the command exits without problems
    assert result.exit_code == 0

    # THEN assert the results folder was created
    assert demux_dir.exists()

    # THEN assert that the sbatch script was created
    assert demux_api.demultiplex_sbatch_path(flow_cell).exists()


def test_demultiplex_all_novaseq(
    cli_runner: testing.CliRunner,
    demultiplexing_context_for_demux: CGConfig,
    tmp_flow_cells_demux_all_directory: Path,
    caplog,
):
    """Test the demultiplex-all command on a directory with newly sequenced NovaSeq6000 flow cells."""
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplexing context with an API and correct structure
    demux_api: DemultiplexingAPI = demultiplexing_context_for_demux.demultiplex_api
    assert demux_api.flow_cells_dir == tmp_flow_cells_demux_all_directory

    # GIVEN sequenced flow cells with their sample sheet in Housekeeper
    for flow_cell_dir in tmp_flow_cells_demux_all_directory.iterdir():
        flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(flow_cell_path=flow_cell_dir)
        add_sample_sheet_path_to_housekeeper(
            flow_cell_directory=flow_cell_dir,
            flow_cell_name=flow_cell.id,
            hk_api=demultiplexing_context_for_demux.housekeeper_api,
        )
        assert demultiplexing_context_for_demux.housekeeper_api.last_version(bundle=flow_cell.id)

    # WHEN running the demultiplex all command
    result: testing.Result = cli_runner.invoke(
        demultiplex_all,
        ["--flow-cells-directory", str(demux_api.flow_cells_dir), "--dry-run"],
        obj=demultiplexing_context_for_demux,
    )

    # THEN assert it exits without problems
    assert result.exit_code == 0

    # THEN assert it found the directory
    assert "Found directory" in caplog.text

    # THEN assert it found a flow cell that is ready for demultiplexing
    assert f"Flow cell {flow_cell.id} is ready for downstream processing" in caplog.text


def test_is_demultiplexing_complete(tmp_flow_cell_directory_bcl2fastq: Path):
    """Tests the is_demultiplexing_complete property of FlowCellDirectoryData."""

    # GIVEN a demultiplexing directory with no demuxcomplete.txt file
    flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(
        flow_cell_path=tmp_flow_cell_directory_bcl2fastq
    )
    assert not flow_cell.is_demultiplexing_complete

    # WHEN creating the demuxcomplete.txt file
    Path(flow_cell.path, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()

    # THEN the property should return true
    assert flow_cell.is_demultiplexing_complete


def test_delete_flow_cell_dry_run_status_db(
    cli_runner: testing.CliRunner,
    tmp_flow_cell_directory_bcl2fastq: Path,
    demultiplex_context: CGConfig,
    tmp_flow_cell_demux_base_path: Path,
    tmp_flow_cell_run_base_path: Path,
    bcl2fastq_flow_cell_id: str,
    caplog,
):
    """Test if logic work - call all true if status_db passed."""
    caplog.set_level(logging.DEBUG)

    demultiplex_context.demultiplex_api.flow_cells_dir = tmp_flow_cell_run_base_path
    demultiplex_context.demultiplex_api.demultiplexed_runs_dir = tmp_flow_cell_demux_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    Path(tmp_flow_cell_demux_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # GIVEN a flow cell to be deleted
    assert bcl2fastq_flow_cell_id in tmp_flow_cell_directory_bcl2fastq.name

    # WHEN deleting a flowcell from status db in dry run mode
    result: testing.Result = cli_runner.invoke(
        delete_flow_cell,
        [
            "-f",
            bcl2fastq_flow_cell_id,
            "--status-db",
            "--dry-run",
            "--yes",
        ],
        obj=demultiplex_context,
    )

    # THEN the code should be executed successfully
    assert result.exit_code == 0

    # THEN it should be notified that it was going to remove all but init-files
    assert (
        f"DeleteDemuxAPI-Housekeeper: Would delete sample sheet files with tag {bcl2fastq_flow_cell_id}"
        in caplog.text
    )
    assert (
        f"DeleteDemuxAPI-Housekeeper: Would delete fastq and spring files related to flow cell {bcl2fastq_flow_cell_id}"
        in caplog.text
    )
    assert f"DeleteDemuxAPI-StatusDB: Would remove {bcl2fastq_flow_cell_id}" in caplog.text

    assert (
        "DeleteDemuxAPI-Hasta: Would have removed the following directory: "
        f"{demultiplex_context.demultiplex_api.demultiplexed_runs_dir / Path(f'some_prefix_1100_{bcl2fastq_flow_cell_id}')}\n"
        f"DeleteDemuxAPI-Hasta: Would have removed the following directory: {demultiplex_context.demultiplex_api.flow_cells_dir / Path(f'some_prefix_1100_{bcl2fastq_flow_cell_id}')}"
    ) in caplog.text
    assert "DeleteDemuxAPI-Init-files: Would have removed" not in caplog.text



"""Module to read or write txt files."""

from pathlib import Path
from typing import List, Union


def read_txt(file_path: Path, read_to_string: bool = False) -> Union[List[str], str]:
    """Read content in a TXT file."""
    with open(file_path, "r") as file:
        if read_to_string:
            return file.read()
        return list(file)



"""This script tests the cli methods to set values of several cases in status-db."""
import logging

import pytest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.set.cases import set_cases
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Family, Sample
from tests.store_helpers import StoreHelpers


@pytest.mark.parametrize("identifier_key", ["original_ticket", "order"])
def test_set_cases_by_sample_identifiers(
    cli_runner: CliRunner,
    base_context: CGConfig,
    identifier_key: str,
    helpers: StoreHelpers,
    caplog: LogCaptureFixture,
    ticket_id: str,
):
    # GIVEN a database with a case with a sample
    base_store: Store = base_context.status_db
    new_sample: Sample = helpers.add_sample(base_store)
    new_sample.original_ticket: str = ticket_id
    new_sample.order: str = "An order"
    case: Family = helpers.add_case(base_store)
    helpers.add_relationship(base_store, sample=new_sample, case=case)
    identifier_value = getattr(new_sample, identifier_key)

    caplog.set_level(logging.INFO)

    # WHEN calling set families with valid sample identifiers
    cli_runner.invoke(
        set_cases,
        ["--sample-identifier", identifier_key, identifier_value],
        obj=base_context,
    )

    # THEN it should name the case to be changed
    assert case.internal_id in caplog.text
    assert case.name in caplog.text



from datetime import datetime
from typing import List

import pytest

from cg.constants import Pipeline
from cg.constants.invoice import CustomerNames
from cg.store import Store
from cg.store.models import Analysis, Sample
from cg.utils.dispatcher import Dispatcher
from tests.store_helpers import StoreHelpers


def test_dispatch_table_generation(
    param1: str = None, param2: str = None, param3: str = None, param4: str = None
):
    """Test that the dispatch table is generated correctly."""

    # Given a list of functions with different parameters
    def test_function_one(param1: str):
        pass

    def test_function_two(param1: str, param2: str):
        pass

    def test_function_three(param1: str, param2: str, param3: str):
        pass

    def test_function_four(param4: str):
        pass

    # When generating the dispatch table
    dispatcher = Dispatcher(
        [test_function_one, test_function_two, test_function_three, test_function_four],
        {"param1": param1, "param2": param2, "param3": param3, "param4": param4},
    )

    expected_table = {
        ("param1",): test_function_one,
        ("param1", "param2"): test_function_two,
        ("param1", "param2", "param3"): test_function_three,
        ("param4",): test_function_four,
    }
    # Then the dispatch table should contain the correct keys and values
    assert dispatcher.dispatch_table.keys() == expected_table.keys()
    for key in expected_table:
        assert dispatcher.dispatch_table[key] == expected_table[key]


def test_call_matching_function(
    a: int = None, b: int = None, c: int = None, x: int = 5, y: int = 4
):
    """Test that the dispatcher can be called with a dictionary that matches one of the functions."""

    # Given two functions
    def foo(a, b, c):
        return a + b + c

    def bar(x, y):
        return x * y

    # When calling the dispatcher with a dictionary that matches one of the functions
    dispatcher = Dispatcher([foo, bar], input_dict={"a": a, "b": b, "c": c, "x": x, "y": y})

    # Then the dispatcher should return the correct result
    assert dispatcher() == 20


def test_call_non_matching_function(a: int = 1, b: int = 2, c: int = 3, x: int = 5, y: int = 4):
    """Test that the dispatcher raises an error when called with a dictionary that does not match any of the functions."""

    # Given two functions
    def foo(a, b, c):
        return a + b + c

    def bar(x, y):
        return x * y

    # When calling the dispatcher with a dictionary that does not match any of the functions
    dispatcher = Dispatcher([foo, bar], input_dict={"a": a, "b": b, "c": c, "x": x, "y": y})

    # Then the dispatcher should raise an error
    with pytest.raises(ValueError):
        dispatcher()


def test_call_for_function_without_parameters():
    """Test that the dispatcher can be called with a function that has no parameters."""

    # GIVEN a function with no parameters
    def foo():
        return 1

    # WHEN calling the dispatcher with a dictionary that has no parameters
    dispatcher = Dispatcher([foo], input_dict={})

    # THEN the dispatcher should return the correct result
    assert dispatcher() == 1


def test_call_for_function_with_same_number_of_parameters(a=1, b=2, c=None, d=None):
    """Test that the dispatcher can be called with a dictionary that has the same number of parameters as the functions."""

    # GIVEN two functions with the same number of parameters
    def foo(a, b):
        return a + b

    def bar(c, d):
        return c * d

    # WHEN calling the dispatcher with a dictionary that has the same number of parameters as the functions
    dispatcher = Dispatcher([foo, bar], input_dict={"a": a, "b": b, "c": c, "d": d})

    # THEN the dispatcher should return the correct result for each function
    assert dispatcher() == 3


def test_call_dictionary_extra_parameters_not_in_functions(
    a: int = 1, b: int = 2, c: int = 3, d: int = None, e: int = None
):
    """Test that the dispatcher can be called with a dictionary that has extra parameters that are not in the functions."""

    # GIVEN two functions with different parameters
    def foo(a, b):
        return a + b

    def bar(c, d):
        return c * d

    # WHEN calling the dispatcher with a dictionary that has extra parameters
    dispatcher = Dispatcher([foo, bar], input_dict={"a": a, "b": b, "c": c, "d": d, "e": e})

    # THEN the dispatcher should return a value error
    with pytest.raises(ValueError):
        dispatcher()


def test_call_with_status_db_functions(
    store: Store,
    helpers: StoreHelpers,
    customer_internal_id: str = CustomerNames.cust001,
    test_subject: str = "test_subject",
    is_tumour: bool = True,
):
    """Test that the dispatcher can be used to call functions in the status db"""

    # GIVEN a database with a customer, a subject and two samples
    helpers.add_sample(store, subject_id=test_subject)
    helpers.add_sample(store, subject_id=test_subject, is_tumour=False)

    # WHEN calling the dispatcher with the customer and subject id
    dispatcher = Dispatcher(
        functions=[
            store.get_samples_by_customer_and_subject_id,
        ],
        input_dict={
            "customer_internal_id": customer_internal_id,
            "subject_id": test_subject,
        },
    )
    # THEN the dispatcher should return the correct samples
    samples: List[Sample] = dispatcher()
    for sample in samples:
        assert sample.customer.internal_id == customer_internal_id
        assert sample.subject_id == test_subject


def test_dispatcher_on_other_functions(
    store: Store,
    helpers: StoreHelpers,
    timestamp_now: datetime,
    timestamp_yesterday: datetime,
    pipeline: Pipeline = Pipeline.MIP_DNA,
    case_internal_id: str = "test_case",
):
    """Test that the dispatcher can be used to call functions in the status db"""

    # GIVEN a database with a case and an analysis
    case = helpers.add_case(store, internal_id=case_internal_id)
    helpers.add_analysis(store, case=case, started_at=timestamp_yesterday, pipeline=pipeline)
    helpers.add_analysis(store, case=case, started_at=timestamp_now, pipeline=Pipeline.FLUFFY)
    helpers.add_analysis(store, case=case, started_at=timestamp_yesterday, pipeline=Pipeline.FLUFFY)

    # WHEN calling the dispatcher with the to get analyses
    function_dispatcher: Dispatcher = Dispatcher(
        functions=[
            store.get_analyses_started_at_before,
            store.get_analyses_for_case_and_pipeline_started_at_before,
            store.get_analyses_for_pipeline_started_at_before,
            store.get_analyses_for_case_started_at_before,
        ],
        input_dict={
            "case_internal_id": case_internal_id,
            "pipeline": pipeline,
            "started_at_before": timestamp_now,
        },
    )
    analyses: List[Analysis] = function_dispatcher()

    # THEN the dispatcher should return the correct analyses
    for analysis in analyses:
        assert analysis
        assert analysis.family.internal_id == case_internal_id
        assert analysis.pipeline == pipeline
        assert analysis.started_at < timestamp_now



""" Module to group PDC related commands """

import logging

from cg.constants.pdc import DSMCParameters
from cg.utils import Process

LOG = logging.getLogger(__name__)

SERVER = "hasta"


class PdcAPI:
    """Group PDC related commands"""

    def __init__(self, binary_path: str = None, dry_run: bool = False):
        self.process: Process = Process(binary=binary_path)
        self.dry_run: bool = dry_run

    def archive_file_to_pdc(self, file_path: str, dry_run: bool = False) -> None:
        """Archive a file by storing it on PDC"""
        command: list = DSMCParameters.ARCHIVE_COMMAND.copy()
        command.append(file_path)
        if not dry_run:
            self.run_dsmc_command(command=command)

    def query_pdc(self, search_pattern: str) -> None:
        """Query PDC based on a given search pattern."""
        command: list = DSMCParameters.QUERY_COMMAND.copy()
        command.append(search_pattern)
        LOG.debug("Starting DSMC command:")
        LOG.debug(f"{self.process.binary} {' '.join(command)}")
        self.process.run_command(parameters=command)

    def retrieve_file_from_pdc(self, file_path: str, target_path: str = None) -> None:
        """Retrieve a file from PDC"""
        command: list = DSMCParameters.RETRIEVE_COMMAND.copy()
        command.append(file_path)
        if target_path:
            command.append(target_path)
        self.run_dsmc_command(command=command)

    def run_dsmc_command(self, command: list) -> None:
        """Runs a DSMC command"""
        LOG.debug("Starting DSMC command:")
        LOG.debug(f"{self.process.binary} {' '.join(command)}")
        self.process.run_command(parameters=command, dry_run=self.dry_run)



import logging
from collections import defaultdict
from typing import List, Tuple

from pydantic import BaseModel, ConfigDict, Field

from cg.apps.demultiplex.sample_sheet.validators import SampleId
from cg.constants.constants import GenomeVersion
from cg.constants.demultiplexing import (
    SampleSheetBcl2FastqSections,
    SampleSheetBCLConvertSections,
)

LOG = logging.getLogger(__name__)


class FlowCellSample(BaseModel):
    """Base class for flow cell samples."""

    lane: int
    sample_id: SampleId
    index: str
    index2: str = ""
    model_config = ConfigDict(populate_by_name=True, extra="ignore")


class FlowCellSampleBcl2Fastq(FlowCellSample):
    """Base class for NovaSeq6000 flow cell samples."""

    flowcell_id: str = Field("", alias=SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value)
    lane: int = Field(..., alias=SampleSheetBcl2FastqSections.Data.LANE.value)
    sample_ref: str = Field(
        GenomeVersion.hg19.value, alias=SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE.value
    )
    index: str = Field(..., alias=SampleSheetBcl2FastqSections.Data.INDEX_1.value)
    index2: str = Field("", alias=SampleSheetBcl2FastqSections.Data.INDEX_2.value)
    sample_name: str = Field(..., alias=SampleSheetBcl2FastqSections.Data.SAMPLE_NAME.value)
    control: str = Field("N", alias=SampleSheetBcl2FastqSections.Data.CONTROL.value)
    recipe: str = Field("R1", alias=SampleSheetBcl2FastqSections.Data.RECIPE.value)
    operator: str = Field("script", alias=SampleSheetBcl2FastqSections.Data.OPERATOR.value)

    sample_id: SampleId = Field(
        ..., alias=SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value
    )
    project: str = Field(
        ..., alias=SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ.value
    )


class FlowCellSampleBCLConvert(FlowCellSample):
    """Class that represents a NovaSeqX flow cell sample."""

    lane: int = Field(..., alias=SampleSheetBCLConvertSections.Data.LANE.value)
    sample_id: SampleId = Field(
        ..., alias=SampleSheetBCLConvertSections.Data.SAMPLE_INTERNAL_ID.value
    )
    index: str = Field(..., alias=SampleSheetBCLConvertSections.Data.INDEX_1.value)
    index2: str = Field("", alias=SampleSheetBCLConvertSections.Data.INDEX_2.value)
    override_cycles: str = Field("", alias=SampleSheetBCLConvertSections.Data.OVERRIDE_CYCLES.value)
    adapter_read_1: str = Field("", alias=SampleSheetBCLConvertSections.Data.ADAPTER_READ_1.value)
    adapter_read_2: str = Field("", alias=SampleSheetBCLConvertSections.Data.ADAPTER_READ_2.value)
    barcode_mismatches_1: int = Field(
        1, alias=SampleSheetBCLConvertSections.Data.BARCODE_MISMATCHES_1.value
    )
    barcode_mismatches_2: int = Field(
        1, alias=SampleSheetBCLConvertSections.Data.BARCODE_MISMATCHES_2.value
    )


class SampleSheet(BaseModel):
    samples: List[FlowCellSample]

    def get_non_pooled_lanes_and_samples(self) -> List[Tuple[int, str]]:
        """Return tuples of non-pooled lane and sample ids."""
        non_pooled_lane_sample_id_pairs: List[Tuple[int, str]] = []
        non_pooled_samples: List[FlowCellSample] = self.get_non_pooled_samples()
        for sample in non_pooled_samples:
            non_pooled_lane_sample_id_pairs.append((sample.lane, sample.sample_id))
        return non_pooled_lane_sample_id_pairs

    def get_non_pooled_samples(self) -> List[FlowCellSample]:
        """Return samples that are sequenced solo in their lane."""
        lane_samples = defaultdict(list)
        for sample in self.samples:
            lane_samples[sample.lane].append(sample)
        return [samples[0] for samples in lane_samples.values() if len(samples) == 1]

    def get_sample_ids(self) -> List[str]:
        """Return ids for samples in sheet."""
        sample_internal_ids: List[str] = []
        for sample in self.samples:
            sample_internal_ids.append(sample.sample_id)
        return list(set(sample_internal_ids))


class SampleSheetBcl2Fastq(SampleSheet):
    samples: List[FlowCellSampleBcl2Fastq]


class SampleSheetBCLConvert(SampleSheet):
    samples: List[FlowCellSampleBCLConvert]
